from collections import namedtuple
import sys
import re
import warnings
import array

import numpy as np

from scipy._lib._util import check_random_state
from scipy.optimize import minimize, differential_evolution, least_squares
import scipy.optimize as sciopt

from refnx.analysis import Objective, Interval, PDF, is_parameter
from refnx._lib import (unique as f_unique, MapWrapper,
                        possibly_open_file, flatten)
from refnx._lib.util import getargspec

from refnx._lib import emcee
from refnx._lib.emcee.state import State

# PTSampler has been forked into a separate package. Try both places
_HAVE_PTSAMPLER = False
PTSampler = type(None)

try:
    from ptemcee.sampler import Sampler as PTSampler
    _HAVE_PTSAMPLER = True
except ImportError:
    warnings.warn("PTSampler (parallel tempering) is not available,"
                  " please install the ptemcee package", ImportWarning)

MCMCResult = namedtuple('MCMCResult', ['name', 'param', 'stderr', 'chain',
                                       'median'])


class CurveFitter(object):
    """
    Analyse a curvefitting system (with MCMC sampling)

    Parameters
    ----------
    objective : refnx.analysis.Objective
        The :class:`refnx.analysis.Objective` to be analysed.
    nwalkers : int, optional
        How many walkers you would like the sampler to have. Must be an
        even number. The more walkers the better.
    ntemps : int or None, optional
        If `ntemps == -1`, then an :class:`emcee.EnsembleSampler` is used
        during the `sample` method.
        Otherwise, or if `ntemps is None` then parallel tempering is
        used with a :class:`ptemcee.sampler.Sampler` object during the `sample`
        method, with `ntemps` specifing the number of temperatures. Can be
        `None`, in which case the `Tmax` keyword argument sets the maximum
        temperature. Parallel Tempering is useful if you expect your
        posterior distribution to be multi-modal.

    mcmc_kws : dict
        Keywords used to create the :class:`emcee.EnsembleSampler` or
        :class:`ptemcee.sampler.Sampler` objects.

    Notes
    -----
    See the documentation at http://dan.iel.fm/emcee/current/api/ for
    further details on what keywords are permitted, and for further
    information on Parallel Tempering. The `pool` and `threads` keywords
    are ignored here. Specification of parallel threading is done with the
    `pool` argument in the `sample` method.
    """
    def __init__(self, objective, nwalkers=200, ntemps=-1, **mcmc_kws):
        """
        Parameters
        ----------
        objective : refnx.analysis.Objective
            The :class:`refnx.analysis.Objective` to be analysed.
        nwalkers : int, optional
            How many walkers you would like the sampler to have. Must be an
            even number. The more walkers the better.
        ntemps : int or None, optional
            If `ntemps == -1`, then an :class:`emcee.EnsembleSampler` is used
            during the `sample` method.
            Otherwise, or if `ntemps is None` then parallel tempering is
            used with a :class:`ptemcee.sampler.Sampler` object during the
            `sample` method, with `ntemps` specifing the number of
            temperatures. Can be `None`, in which case the `Tmax` keyword
            argument sets the maximum temperature. Parallel Tempering is
            useful if you expect your posterior distribution to be multi-modal.
        mcmc_kws : dict
            Keywords used to create the :class:`emcee.EnsembleSampler` or
            :class:`ptemcee.sampler.PTSampler` objects.

        Notes
        -----
        See the documentation at http://dan.iel.fm/emcee/current/api/ for
        further details on what keywords are permitted. The `pool` and
        keyword is ignored here. Specification of parallel threading is done
        with the `pool` argument in the `sample` method.
        To use parallel tempering you will need to install the
        :package:`ptemcee` package.
        """
        self.objective = objective
        self._varying_parameters = []
        self.__var_id = []

        self.mcmc_kws = {}
        if mcmc_kws is not None:
            self.mcmc_kws.update(mcmc_kws)

        if 'pool' in self.mcmc_kws:
            self.mcmc_kws.pop('pool')
        if 'threads' in self.mcmc_kws:
            self.mcmc_kws.pop('threads')

        self._nwalkers = nwalkers
        self._ntemps = ntemps
        self.make_sampler()
        self._state = None

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__var_id = [id(obj) for obj
                         in self.objective.varying_parameters()]

    @property
    def nvary(self):
        return len(self._varying_parameters)

    def __repr__(self):
        # attempt to get a minimum repr for a CurveFitter. However,
        # it has so much state when the sampling has been done, that
        # will be ignored.
        d = {'objective': self.objective, '_nwalkers': self._nwalkers,
             '_ntemps': self._ntemps, 'mcmc_kws': self.mcmc_kws}
        return ("CurveFitter({objective!r},"
                " nwalkers={_nwalkers},"
                " ntemps={_ntemps},"
                " {mcmc_kws!r})".format(**d))

    def make_sampler(self):
        """
        Make the samplers for the Objective.

        Use this method if the number of varying parameters changes.
        """
        self._varying_parameters = self.objective.varying_parameters()
        self.__var_id = [id(obj) for obj in self._varying_parameters]

        if not self.nvary:
            raise ValueError("No parameters are being fitted")

        if self._ntemps == -1:
            self.sampler = emcee.EnsembleSampler(self._nwalkers,
                                                 self.nvary,
                                                 self.objective.logpost,
                                                 **self.mcmc_kws)
        # Parallel Tempering was requested.
        else:
            if not _HAVE_PTSAMPLER:
                raise RuntimeError("You need to install the 'ptemcee' package"
                                   " to use parallel tempering")

            sig = {'ntemps': self._ntemps,
                   'nwalkers': self._nwalkers,
                   'dim': self.nvary,
                   'logl': self.objective.logl,
                   'logp': self.objective.logp
                   }
            sig.update(self.mcmc_kws)
            self.sampler = PTSampler(**sig)

            # construction of the PTSampler creates an ntemps attribute.
            # If it was constructed with ntemps = None, then ntemps will
            # be an integer.
            self._ntemps = self.sampler.ntemps

        self._state = None

    def _check_vars_unchanged(self):
        """
        Keep track of whether the varying parameters have changed after
        construction of CurveFitter object.

        """
        var_ids = [id(obj) for obj in self.objective.varying_parameters()]
        if not(np.array_equal(var_ids, self.__var_id)):
            raise RuntimeError("The Objective.varying_parameters() have"
                               " changed since the CurveFitter was created."
                               " To keep on using the CurveFitter call"
                               " the CurveFitter.make_samplers() method.")

    def initialise(self, pos='covar'):
        """
        Initialise the emcee walkers.

        Parameters
        ----------
        pos : str or np.ndarray
            Method for initialising the emcee walkers. One of:

            - 'covar', use the estimated covariance of the system.
            - 'jitter', add a small amount of gaussian noise to each parameter
            - 'prior', sample random locations from the prior
            - pos, an array that specifies a snapshot of the walkers. Has shape
                `(nwalkers, ndim)`, or `(ntemps, nwalkers, ndim)` if parallel
                 tempering is employed. You can also provide a previously
                 created chain.
        """
        nwalkers = self._nwalkers
        nvary = self.nvary

        # account for parallel tempering
        _ntemps = self._ntemps

        # If you're not doing parallel tempering, temporarily set the number of
        # temperatures to be created to 1, thereby producing initial positions
        # of (1, nwalkers, nvary), this first dimension should be removed at
        # the end of the method
        if self._ntemps == -1:
            _ntemps = 1

        # position is specified with array (no parallel tempering)
        if (isinstance(pos, np.ndarray) and
                self._ntemps == -1 and
                pos.shape == (nwalkers, nvary)):
            init_walkers = np.copy(pos)[np.newaxis]

        # position is specified with array (with parallel tempering)
        elif (isinstance(pos, np.ndarray) and
              self._ntemps > -1 and
              pos.shape == (_ntemps, nwalkers, nvary)):
            init_walkers = np.copy(pos)

        # position is specified with existing chain
        elif isinstance(pos, np.ndarray):
            self.initialise_with_chain(pos)
            return

        # position is to be created from covariance matrix
        elif pos == 'covar':
            p0 = np.array(self._varying_parameters)
            cov = self.objective.covar()
            init_walkers = np.random.multivariate_normal(
                np.atleast_1d(p0),
                np.atleast_2d(cov),
                size=(_ntemps, nwalkers))

        # position is specified by jittering the parameters with gaussian noise
        elif pos == 'jitter':
            var_arr = np.array(self._varying_parameters)
            pos = 1 + np.random.randn(_ntemps,
                                      nwalkers,
                                      nvary) * 1.e-4
            pos *= var_arr
            init_walkers = pos

        # use the prior to initialise position
        elif pos == 'prior':
            arr = np.zeros((_ntemps, nwalkers, nvary))

            for i, param in enumerate(self._varying_parameters):
                # bounds are not a closed interval, just jitter it.
                if (isinstance(param.bounds, Interval) and
                        not param.bounds._closed_bounds):
                    vals = ((1 + np.random.randn(_ntemps, nwalkers) * 1.e-1) *
                            param.value)
                    arr[..., i] = vals
                else:
                    arr[..., i] = param.bounds.rvs(size=(_ntemps, nwalkers))

            init_walkers = arr

        else:
            raise RuntimeError("Didn't use any known method for "
                               "CurveFitter.initialise")

        # if you're not doing parallel tempering then remove the first
        # dimension
        if self._ntemps == -1:
            init_walkers = init_walkers[0]

        # now validate initialisation, ensuring all init pos have finite
        # logpost
        for i, param in enumerate(self._varying_parameters):
            init_walkers[..., i] = param.valid(init_walkers[..., i])

        self._state = State(init_walkers)

        # finally reset the sampler to reset the chain
        # you have to do this at the end, not at the start because resetting
        # makes self.sampler.chain == None and the PTsampler creation doesn't
        # work
        self.sampler.reset()

    def initialise_with_chain(self, chain):
        """
        Initialise sampler with a pre-existing chain

        Parameters
        ----------
        chain : array
            Array of size `(steps, ntemps, nwalkers, ndim)` or
            `(steps, nwalkers, ndim)`, containing a chain from a previous
            sampling run.
        """
        # we should be left with (nwalkers, ndim) or (ntemp, nwalkers, ndim)

        if self._ntemps == -1:
            required_shape = (self._nwalkers, self.nvary)
        else:
            required_shape = (self._ntemps, self._nwalkers, self.nvary)

        chain_shape = chain.shape[1:]

        # if the shapes are the same, then we can initialise
        if required_shape == chain_shape:
            self.initialise(pos=chain[-1])
        else:
            raise ValueError("You tried to initialise with a chain, but it was"
                             " the wrong shape")

    @property
    def chain(self):
        """
        MCMC chain belonging to CurveFitter.sampler

        Returns
        -------
        chain : array
            The MCMC chain with shape `(steps, nwalkers, ndim)` or
            `(steps, ntemps, nwalkers, ndim)`.

        Notes
        -----
        The chain returned here has swapped axes compared to the
        `PTSampler.chain` and `EnsembleSampler.chain` attributes
        """
        if isinstance(self.sampler, PTSampler):
            return np.transpose(self.sampler.chain, axes=(2, 0, 1, 3))

        return self.sampler.get_chain()

    @property
    def logpost(self):
        """
        Log-probability for each of the entries in `self.chain`
        """
        if isinstance(self.sampler, PTSampler):
            return np.transpose(self.sampler.logprobability, axes=(2, 0, 1))

        return self.sampler.get_log_prob()

    def reset(self):
        """
        Reset the sampled chain.

        Typically used on a sampler after a burn-in period.
        """
        self.sampler.reset()

    def acf(self, nburn=0, nthin=1):
        """
        Calculate the autocorrelation function

        Returns
        -------
        acfs : np.ndarray
            The autocorrelation function, acfs.shape=(lags, nvary)
        """
        return autocorrelation_chain(self.chain, nburn=nburn, nthin=nthin)

    def sample(self, steps, nthin=1, random_state=None, f=None, callback=None,
               verbose=True, pool=-1):
        """
        Performs sampling from the objective.

        Parameters
        ----------
        steps : int
            Collect `steps` samples into the chain. The sampler will run a
            total of `steps * nthin` moves.
        nthin : int, optional
            Each chain sample is separated by `nthin` iterations.
        random_state : int or `np.random.RandomState`, optional
            If `random_state` is an int, a new `np.random.RandomState` instance
            is used, seeded with `random_state`.
            If `random_state` is already a `np.random.RandomState` instance,
            then that `np.random.RandomState` instance is used. Specify
            `random_state` for repeatable sampling
        f : file-like or str
            File to incrementally save chain progress to. Each row in the file
            is a flattened array of size `(nwalkers, ndim)` or
            `(ntemps, nwalkers, ndim)`. There are `steps` rows in the
            file.
        callback : callable
            callback function to be called at each iteration step. Has the
            signature `callback(coords, logprob)`.
        verbose : bool, optional
            Gives updates on the sampling progress
        pool : int or map-like object, optional
            If `pool` is an `int` then it specifies the number of threads to
            use for parallelization. If `pool == -1`, then all CPU's are used.
            If pool is a map-like callable that follows the same calling
            sequence as the built-in map function, then this pool is used for
            parallelisation.

        Notes
        -----
        Please see :class:`emcee.EnsembleSampler` for its detailed behaviour.

        >>> # we'll burn the first 500 steps
        >>> fitter.sample(500)
        >>> # after you've run those, then discard them by resetting the
        >>> # sampler.
        >>> fitter.sampler.reset()
        >>> # Now collect 40 steps, each step separated by 50 sampler
        >>> # generations.
        >>> fitter.sample(40, nthin=50)

        One can also burn and thin in `Curvefitter.process_chain`.
        """
        self._check_vars_unchanged()

        if self._state is None:
            self.initialise()

        self.__pt_iterations = 0
        if isinstance(self.sampler, PTSampler):
            steps *= nthin

        # for saving progress to file
        def _callback_wrapper(state, h=None):
            if callback is not None:
                callback(state.coords, state.log_prob)

            if h is not None:
                # if you're parallel tempering, then you only
                # want to save every nthin
                if isinstance(self.sampler, PTSampler):
                    self.__pt_iterations += 1
                    if self.__pt_iterations % nthin:
                        return None

                h.write(' '.join(map(str, state.coords.ravel())))
                h.write('\n')

        # set the random state of the sampler
        # normally one could give this as an argument to the sample method
        # but PTSampler didn't historically accept that...
        if random_state is not None:
            rstate0 = check_random_state(random_state).get_state()
            self._state.random_state = rstate0
            if isinstance(self.sampler, PTSampler):
                self.sampler._random = rstate0

        # remove chains from each of the parameters because they slow down
        # pickling but only if they are parameter objects.
        flat_params = f_unique(flatten(self.objective.parameters))
        flat_params = [param for param in flat_params if is_parameter(param)]
        # zero out all the old parameter stderrs
        for param in flat_params:
            param.stderr = None
            param.chain = None

        # make sure the checkpoint file exists
        if f is not None:
            with possibly_open_file(f, 'w') as h:
                # write the shape of each step of the chain
                h.write('# ')
                shape = self._state.coords.shape
                h.write(', '.join(map(str, shape)))
                h.write('\n')

        # using context manager means we kill off zombie pool objects
        # but does mean that the pool has to be specified each time.
        with MapWrapper(pool) as g, possibly_open_file(f, 'a') as h:
            # if you're not creating more than 1 thread, then don't bother with
            # a pool.
            if pool == 1:
                self.sampler.pool = None
            else:
                self.sampler.pool = g

            # these kwargs are provided to the sampler.sample method
            kwargs = {'iterations': steps,
                      'thin': nthin}

            # new emcee arguments
            sampler_args = getargspec(self.sampler.sample).args
            if 'progress' in sampler_args and verbose:
                kwargs['progress'] = True
                verbose = False

            if 'thin_by' in sampler_args:
                kwargs['thin_by'] = nthin
                kwargs.pop('thin', 0)

            # ptemcee returns coords, logpost
            # emcee returns a State object
            if isinstance(self.sampler, PTSampler):
                for result in self.sampler.sample(self._state.coords,
                                                  **kwargs):
                    self._state = State(result[0],
                                        log_prob=result[1] + result[2],
                                        random_state=self.sampler._random)
                    _callback_wrapper(self._state, h=h)
            else:
                for state in self.sampler.sample(self._state,
                                                 **kwargs):
                    self._state = state
                    _callback_wrapper(state, h=h)

        self.sampler.pool = None

        # finish off the progress bar
        if verbose:
            sys.stdout.write("\n")

        # sets parameter value and stderr
        return process_chain(self.objective, self.chain)

    def fit(self, method='L-BFGS-B', target='nll', **kws):
        """
        Obtain the maximum log-likelihood, or log-posterior, estimate (mode)
        of the objective. Maximising the log-likelihood is equivalent to
        minimising chi2 in a least squares fit.

        Parameters
        ----------
        method : str
            which method to use for the optimisation. One of:

            - `'least_squares'`: `scipy.optimize.least_squares`.
            - `'L-BFGS-B'`: L-BFGS-B
            - `'differential_evolution'`:
                `scipy.optimize.differential_evolution`
            - `'dual_annealing'`:
                `scipy.optimize.dual_annealing` (SciPy >= 1.2.0)
            - `'shgo'`: `scipy.optimize.shgo` (SciPy >= 1.2.0)

            You can also choose many of the minimizers from
            ``scipy.optimize.minimize``.

        target : {'nll', 'nlpost'}, optional
            Minimize the negative log-likelihood (`'nll'`) or the negative
            log-posterior (`'nlpost'`). This is equivalent to maximising the
            likelihood or posterior probabilities respectively.
            Maximising the likelihood is equivalent to minimising chi^2 in a
            least-squares fit.
            This option only applies to the `differential_evolution`, `shgo`,
            `dual_annealing` or `L-BFGS-B` methods.
            These optimisers require lower and upper (box) bounds for each
            parameter. If the `Bounds` on a parameter are not an `Interval`,
            but a `PDF` specifying a statistical distribution, then the lower
            and upper bounds are approximated as
            ``PDF.rv.ppf([0.005, 0.995])``, covering 99 % of the statistical
            distribution.

        kws : dict
            Additional arguments are passed to the underlying minimization
            method.

        Returns
        -------
        result, covar : OptimizeResult, np.ndarray
            `result.x` contains the best fit parameters
            `result.covar` is the covariance matrix for the fit.
            `result.stderr` is the uncertainties on each of the fit parameters.

        Notes
        -----
        If the `objective` supplies a `residuals` method then `least_squares`
        can be used. Otherwise the `nll` method of the `objective` is
        minimised. Use this method just before a sampling run.
        If `self.objective.parameters` is a `Parameters` instance, then each
        of the varying parameters has its value updated by the fit, and each
        `Parameter` has a `stderr` attribute which represents the uncertainty
        on the fit parameter.

        The use of `dual annealing` and `shgo` requires that `scipy >= 1.2.0`
        be installed.

        """
        _varying_parameters = self.objective.varying_parameters()
        init_pars = np.array(_varying_parameters)

        _min_kws = {}
        _min_kws.update(kws)
        _bounds = bounds_list(self.objective.varying_parameters())
        _min_kws['bounds'] = _bounds

        cost = self.objective.nll
        if target == 'nlpost':
            cost = self.objective.nlpost

        # least_squares Trust Region Reflective by default
        if method == 'least_squares':
            b = np.array(_bounds)
            _min_kws['bounds'] = (b[..., 0], b[..., 1])

            # least_squares doesn't have a callback
            if 'callback' in _min_kws:
                _min_kws.pop('callback')

            res = least_squares(self.objective.residuals,
                                init_pars,
                                **_min_kws)
        # differential_evolution, dual_annealing, shgo require lower and upper
        # bounds
        elif method in ['differential_evolution', 'dual_annealing', 'shgo']:
            mini = getattr(sciopt, method)
            res = mini(cost, **_min_kws)
        else:
            # otherwise stick it to minimizer. Default being L-BFGS-B
            _min_kws['method'] = method
            _min_kws['bounds'] = _bounds
            res = minimize(cost, init_pars, **_min_kws)

        # OptimizeResult.success may not be present (dual annealing)
        if hasattr(res, 'success') and res.success:
            self.objective.setp(res.x)

            # Covariance matrix estimation
            covar = self.objective.covar()
            errors = np.sqrt(np.diag(covar))
            res['covar'] = covar
            res['stderr'] = errors

            # check if the parameters are all Parameter instances.
            flat_params = list(f_unique(flatten(
                self.objective.parameters)))
            if np.all([is_parameter(param) for param in flat_params]):
                # zero out all the old parameter stderrs
                for param in flat_params:
                    param.stderr = None
                    param.chain = None

                for i, param in enumerate(_varying_parameters):
                    param.stderr = errors[i]

            # need to touch up the output to check we leave
            # parameters as we found them
            self.objective.setp(res.x)

        return res


def load_chain(f):
    """
    Loads a chain from disk. Does not change the state of a CurveFitter
    object.

    Parameters
    ----------
    f : str or file-like
        File containing the chain.

    Returns
    -------
    chain : array
        The loaded chain - `(nsteps, nwalkers, ndim)` or
        `(nsteps, ntemps, nwalkers, ndim)`
    """
    with possibly_open_file(f, 'r') as g:
        # read header
        header = g.readline()
        expr = re.compile(r"(\d+)")
        matches = expr.findall(header)
        if matches:
            if len(matches) == 3:
                ntemps, nwalkers, ndim = map(int, matches)
                chain_size = ntemps * nwalkers * ndim
            elif len(matches) == 2:
                ntemps = None
                nwalkers, ndim = map(int, matches)
                chain_size = nwalkers * ndim
        else:
            raise ValueError("Couldn't read header line of chain file")

        # make an array that's the appropriate size
        read_arr = array.array("d")

        for i, l in enumerate(g, 1):
            read_arr.extend(np.fromstring(l,
                                          dtype=float,
                                          count=chain_size,
                                          sep=' '))

        chain = np.frombuffer(read_arr, dtype=np.float, count=len(read_arr))

        if ntemps is not None:
            chain = np.reshape(chain, (i, ntemps, nwalkers, ndim))
        else:
            chain = np.reshape(chain, (i, nwalkers, ndim))

        return chain


def process_chain(objective, chain, nburn=0, nthin=1, flatchain=False):
    """
    Process the chain produced by a sampler for a given Objective

    Parameters
    ----------
    objective : refnx.analysis.Objective
        The Objective function that the Posterior was sampled for
    chain : array
        The MCMC chain
    nburn : int, optional
        discard this many steps from the start of the chain
    nthin : int, optional
        only accept every `nthin` samples from the chain
    flatchain : bool, optional
        collapse the walkers down into a single dimension.

    Returns
    -------
    [(param, stderr, chain)] : list
        List of (param, stderr, chain) tuples.
        If `isinstance(objective.parameters, Parameters)` then `param` is a
        `Parameter` instance. `param.value`, `param.stderr` and
        `param.chain` will contain the median, stderr and chain samples,
        respectively. Otherwise `param` will be a float representing the
        median of the chain samples.
        `stderr` is the half width of the [15.87, 84.13] spread (similar to
        standard deviation) and `chain` is an array containing the MCMC
        samples for that parameter.

    Notes
    -----
    The chain should have the shape `(iterations, nwalkers, nvary)` or
    `(iterations, ntemps, nwalkers, nvary)` if parallel tempering was
    employed.
    The burned and thinned chain is created via:
    `chain[nburn::nthin]`.
    Note, if parallel tempering is employed, then only the lowest temperature
    of the parallel tempering chain is processed and returned as it
    corresponds to the (lowest energy) target distribution.
    If `flatten is True` then the burned/thinned chain is reshaped and
    `arr.reshape(-1, nvary)` is returned.
    This function has the effect of setting the parameter stderr's.
    """
    chain = chain[nburn::nthin]
    shape = chain.shape
    nvary = shape[-1]

    # nwalkers = shape[1]
    if len(shape) == 4:
        ntemps = shape[1]
    elif len(shape) == 3:
        ntemps = -1

    if ntemps != -1:
        # PTSampler, we require the target distribution in the first row.
        chain = chain[:, 0]

    _flatchain = chain.reshape((-1, nvary))
    if flatchain:
        chain = _flatchain

    flat_params = list(f_unique(flatten(objective.parameters)))
    varying_parameters = objective.varying_parameters()

    # set the stderr of each of the Parameters
    result_list = []
    if np.all([is_parameter(param) for param in flat_params]):
        # zero out all the old parameter stderrs
        for param in flat_params:
            param.stderr = None
            param.chain = None

        # do the error calcn for the varying parameters and set the chain
        quantiles = np.percentile(_flatchain, [15.87, 50, 84.13], axis=0)
        for i, param in enumerate(varying_parameters):
            std_l, median, std_u = quantiles[:, i]
            param.value = median
            param.stderr = 0.5 * (std_u - std_l)

            # copy in the chain
            param.chain = np.copy(chain[..., i])
            res = MCMCResult(name=param.name, param=param,
                             median=param.value, stderr=param.stderr,
                             chain=param.chain)
            result_list.append(res)

        fitted_values = np.array(varying_parameters)

        # give each constrained param a chain (to be reshaped later)
        constrained_params = [param for param in flat_params
                              if param.constraint is not None]

        for constrain_param in constrained_params:
            constrain_param.chain = np.empty(chain.shape[:-1], float)

        # now iterate through the varying parameters, set the values, thereby
        # setting the constraint value
        if len(constrained_params):
            for index in np.ndindex(chain.shape[:-1]):
                # iterate over parameter vectors
                pvals = chain[index]
                objective.setp(pvals)

                for constrain_param in constrained_params:
                    constrain_param.chain[index] = constrain_param.value

            for constrain_param in constrained_params:
                quantiles = np.percentile(constrain_param.chain,
                                          [15.87, 50, 84.13])

                std_l, median, std_u = quantiles
                constrain_param.value = median
                constrain_param.stderr = 0.5 * (std_u - std_l)

        # now reset fitted parameter values (they would've been changed by
        # constraints calculations
        objective.setp(fitted_values)

    # the parameter set are not Parameter objects, an array was probably
    # being used with BaseObjective.
    else:
        for i in range(nvary):
            c = np.copy(chain[..., i])
            median, stderr = uncertainty_from_chain(c)
            res = MCMCResult(name='', param=median,
                             median=median, stderr=stderr,
                             chain=c)
            result_list.append(res)

    return result_list


def uncertainty_from_chain(chain):
    """
    Calculates the median and uncertainty of MC samples.

    Parameters
    ----------
    chain : array-like

    Returns
    -------
    median, stderr : float, float
        `median` of the chain samples. `stderr` is half the width of the
        [15.87, 84.13] spread.
    """
    flatchain = chain.flatten()
    std_l, median, std_u = np.percentile(flatchain, [15.87, 50, 84.13])
    return median, 0.5 * (std_u - std_l)


def autocorrelation_chain(chain, nburn=0, nthin=1):
    """
    Calculate the autocorrelation function

    Parameters
    ----------
    chain : np.ndarray
        The MCMC chain - `(nsteps, nwalkers, ndim)` or
        `(nsteps, ntemps, nwalkers, ndim)`

    Returns
    -------
    acfs : np.ndarray
        The autocorrelation function, acfs.shape=(lags, nvary)
    """
    lchain = chain
    # parallel tempered chain
    if len(chain.shape) == 4:
        lchain = lchain[:, 0]

    lchain = lchain[nburn::nthin]
    # (iterations, walkers, vary) -> (vary, walkers, iterations)
    lchain = np.swapaxes(lchain, 0, 2)
    shape = lchain.shape[:-1]

    acfs = np.zeros_like(lchain)

    # iterate over each parameter/walker
    for index in np.ndindex(*shape):
        s = _function_1d(lchain[index])
        acfs[index] = s

    # now average over walkers
    acfs = np.mean(acfs, axis=1)
    return np.transpose(acfs)


def bounds_list(parameters):
    """
    Approximates interval bounds for a parameter set.

    Parameters
    ----------
    parameters : sequence
        A sequence containing individual parameters

    Returns
    -------
    bounds: tuple
        ``(min, max)`` pairs that define the finite lower and upper bounds
        every element in ``parameters``.

    If the `Bounds` applied by a parameter are a `PDF` instance then the upper
    and lower bound are approximated by ``PDF.rv.ppf([0.005, 0.995])``, which
    covers 99% of the statistical distribution.
    """
    bounds = []
    for param in parameters:
        if (hasattr(param, 'bounds') and
                isinstance(param.bounds, Interval)):
            bnd = param.bounds
            bounds.append((bnd.lb, bnd.ub))
        elif (hasattr(param, 'bounds') and isinstance(param.bounds, PDF) and
              hasattr(param.bounds.rv, 'ppf')):
            bounds.append(param.bounds.rv.ppf([0.005, 0.995]))
        else:
            # We can't handle this bound
            bounds.append((-np.inf, np.inf))

    return bounds


# Following code is for autocorrelation analysis of chains and is taken from
# emcee.autocorr
def _next_pow_two(n):
    """Returns the next power of two greater than or equal to `n`"""
    i = 1
    while i < n:
        i = i << 1
    return i


def _function_1d(x):
    """Estimate the normalized autocorrelation function of a 1-D series

    Args:
        x: The series as a 1-D numpy array.

    Returns:
        array: The autocorrelation function of the time series.

    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = _next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[:len(x)].real
    acf /= acf[0]
    return acf



import sys


def gui():
    from PyQt5 import QtGui, QtWidgets, QtCore
    from refnx.reflect._app.view import MotofitMainWindow
    from refnx.reflect._app import resources_rc

    app = QtWidgets.QApplication(sys.argv)

    # used to make sure that only decimal point is used for
    # entering floats (3.1), rejecting commas (3,1).
    # This reduces confusion for international users
    lo = QtCore.QLocale(QtCore.QLocale.C)
    lo.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
    QtCore.QLocale.setDefault(lo)

    app.setWindowIcon(QtGui.QIcon(':icons/scattering.png'))
    myapp = MotofitMainWindow()

    fnt = QtGui.QFont('Arial')
    fnt.setPixelSize(12)
    app.setFont(fnt)

    myapp.show()
    myapp.raise_()

    v = app.exec_()
    return v


def main():
    sys.exit(gui())


__all__ = [gui, main]



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["Model"]

from collections import namedtuple

Model = namedtuple(
    "Model",
    ("log_prob_fn", "compute_log_prob_fn", "map_fn", "random")
)



import io
import os
import os.path
import glob
import argparse
import re
import shutil
from time import gmtime, strftime
import string
import warnings
from contextlib import contextmanager

from scipy.optimize import leastsq, curve_fit
from scipy.stats import t
import pandas as pd
import numpy as np
import h5py

from refnx.reduce.peak_utils import peak_finder, centroid
import refnx.util.general as general
from refnx.util.general import resolution_double_chopper, _dict_compare
import refnx.util.ErrorProp as EP
from refnx.reduce.parabolic_motion import (find_trajectory, y_deflection,
                                           parabola)
from refnx.reduce.event import (events, process_event_stream,
                                framebins_to_frames)
from refnx.reduce.rebin import rebin, rebin_along_axis
from refnx._lib import possibly_open_file


# detector y pixel spacing in mm per pixel
Y_PIXEL_SPACING = 1.177

disc_openings = (60., 10., 25., 60.)
O_C1d, O_C2d, O_C3d, O_C4d = disc_openings
O_C1, O_C2, O_C3, O_C4 = np.radians(disc_openings)

DISCRADIUS = 350.
EXTENT_MULT = 2
PIXEL_OFFSET = 1

spectrum_template = """<?xml version="1.0"?>
<REFroot xmlns="">
<REFentry time="$time">
<Title>$title</Title>
<REFdata axes="lambda" rank="1" type="POINT"\
 spin="UNPOLARISED" dim="$n_spectra">
<Run filename="$runnumber"/>
<R uncertainty="dR">$r</R>
<lambda uncertainty="dlambda" units="1/A">$lmda</lambda>
<dR type="SD">$dr</dR>
<dlambda type="_FWHM" units="1/A">$dlmda</dlambda>
</REFdata>
</REFentry>
</REFroot>"""


def catalogue(start, stop, data_folder=None):
    """
    Extract interesting information from Platypus NeXUS files.

    Parameters
    ----------
    start : int
        start cataloguing from this run number.
    stop : int
        stop cataloguing at this run number
    data_folder : str, optional
        path specifying location of NeXUS files

    Returns
    -------
    catalog : pd.DataFrame
        Dataframe containing interesting parameters from Platypus Nexus files
    """
    info = ['filename', 'end_time', 'sample_name', 'ss1vg', 'ss2vg', 'ss3vg',
            'ss4vg', 'omega', 'twotheta', 'total_counts', 'bm1_counts', 'time',
            'daq_dirname', 'start_time']
    run_number = []
    d = {key: [] for key in info}

    if data_folder is None:
        data_folder = '.'

    files = glob.glob(os.path.join(data_folder, '*.nx.hdf'))
    files.sort()
    files = [file for file in files
             if datafile_number(file) in range(start, stop + 1)]

    for idx, file in enumerate(files):
        pn = PlatypusNexus(file)

        cat = pn.cat.cat
        run_number.append(idx)

        for key, val in d.items():
            data = cat[key]
            if np.size(data) > 1 or type(data) is np.ndarray:
                data = data[0]
            if type(data) is bytes:
                data = data.decode()

            d[key].append(data)

    df = pd.DataFrame(d, index=run_number, columns=info)

    return df


class Catalogue(object):
    """
    Extract relevant parts of a NeXus file for reflectometry reduction
    """
    def __init__(self, h5d):
        """
        Extract relevant parts of a NeXus file for reflectometry reduction
        Access information via dict access, e.g. cat['detector'].

        Parameters
        ----------
        h5d - HDF5 file handle
        """
        d = {}
        file_path = os.path.realpath(h5d.filename)
        d['path'] = os.path.dirname(file_path)
        d['filename'] = h5d.filename
        d['end_time'] = h5d['entry1/end_time'][0]

        try:
            d['start_time'] = (
                h5d['entry1/instrument/detector/start_time'][:])
        except KeyError:
            # start times don't exist in this file
            d['start_time'] = None

        d['sample_name'] = h5d['entry1/sample/name'][:]
        d['ss1vg'] = h5d['entry1/instrument/slits/first/vertical/gap'][:]
        d['ss2vg'] = h5d['entry1/instrument/slits/second/vertical/gap'][:]
        d['ss3vg'] = h5d['entry1/instrument/slits/third/vertical/gap'][:]
        d['ss4vg'] = h5d['entry1/instrument/slits/fourth/vertical/gap'][:]
        d['ss1hg'] = h5d['entry1/instrument/slits/first/horizontal/gap'][:]
        d['ss2hg'] = h5d['entry1/instrument/slits/second/horizontal/gap'][:]
        d['ss3hg'] = h5d['entry1/instrument/slits/third/horizontal/gap'][:]
        d['ss4hg'] = h5d['entry1/instrument/slits/fourth/horizontal/gap'][:]

        d['omega'] = h5d['entry1/instrument/parameters/omega'][:]
        d['twotheta'] = h5d['entry1/instrument/parameters/twotheta'][:]

        d['detector'] = h5d['entry1/data/hmm'][:]
        d['sth'] = h5d['entry1/sample/sth'][:]
        d['bm1_counts'] = h5d['entry1/monitor/bm1_counts'][:]
        d['total_counts'] = h5d['entry1/instrument/detector/total_counts'][:]

        d['time'] = h5d['entry1/instrument/detector/time'][:]
        d['mode'] = h5d['entry1/instrument/parameters/mode'][0].decode()

        try:
            event_directory_name = h5d[
                'entry1/instrument/detector/daq_dirname'][0]
            d['daq_dirname'] = event_directory_name.decode()
        except KeyError:
            # daq_dirname doesn't exist in this file
            d['daq_dirname'] = None

        d['t_bins'] = h5d['entry1/data/time_of_flight'][:].astype('float64')
        d['x_bins'] = h5d['entry1/data/x_bin'][:]
        d['y_bins'] = h5d['entry1/data/y_bin'][:]

        master, slave, frequency, phase = self._chopper_values(h5d)
        d['master'] = master
        d['slave'] = slave
        d['frequency'] = frequency
        d['phase'] = phase
        d['chopper2_distance'] = h5d[
            'entry1/instrument/parameters/chopper2_distance'][:]
        d['chopper3_distance'] = h5d[
            'entry1/instrument/parameters/chopper3_distance'][:]
        d['chopper4_distance'] = h5d[
            'entry1/instrument/parameters/chopper4_distance'][:]
        d['chopper1_phase_offset'] = h5d[
            'entry1/instrument/parameters/chopper1_phase_offset'][:]
        d['chopper2_phase_offset'] = h5d[
            'entry1/instrument/parameters/chopper2_phase_offset'][:]
        d['chopper3_phase_offset'] = h5d[
            'entry1/instrument/parameters/chopper3_phase_offset'][:]
        d['chopper4_phase_offset'] = h5d[
            'entry1/instrument/parameters/chopper4_phase_offset'][:]
        d['guide1_distance'] = h5d[
            'entry1/instrument/parameters/guide1_distance'][:]
        d['guide2_distance'] = h5d[
            'entry1/instrument/parameters/guide2_distance'][:]
        d['sample_distance'] = h5d[
            'entry1/instrument/parameters/sample_distance'][:]
        d['slit2_distance'] = h5d[
            'entry1/instrument/parameters/slit2_distance'][:]
        d['slit3_distance'] = h5d[
            'entry1/instrument/parameters/slit3_distance'][:]
        d['collimation_distance'] = d['slit3_distance'] - d['slit2_distance']
        d['dy'] = h5d[
            'entry1/instrument/detector/longitudinal_translation'][:]
        d['dz'] = h5d[
            'entry1/instrument/detector/vertical_translation'][:]
        d['original_file_name'] = h5d['entry1/experiment/file_name']

        d['scan_axis_name'] = (
            h5d['entry1/data/hmm'].attrs['axes'].decode('utf8').split(':')[0])
        d['scan_axis'] = h5d['entry1/data/%s' % d['scan_axis_name']][:]

        # TODO put HDF file y pixel spacing in here.
        self.cat = d

    def __getattr__(self, item):
        return self.cat[item]

    @property
    def datafile_number(self):
        return datafile_number(self.filename)

    def _chopper_values(self, h5data):
        """
        Obtains chopper settings from NeXUS file

        Parameters
        ----------
        h5data : HDF5 NeXUS file
            datafile,

        Returns
        -------
        master, slave, frequency, phase : float, float, float, float
        """
        chopper1_speed = h5data['entry1/instrument/disk_chopper/ch1speed']
        chopper2_speed = h5data['entry1/instrument/disk_chopper/ch2speed']
        chopper3_speed = h5data['entry1/instrument/disk_chopper/ch3speed']
        chopper4_speed = h5data['entry1/instrument/disk_chopper/ch4speed']
        ch2phase = h5data['entry1/instrument/disk_chopper/ch2phase']
        ch3phase = h5data['entry1/instrument/disk_chopper/ch3phase']
        ch4phase = h5data['entry1/instrument/disk_chopper/ch4phase']

        if ('entry1/instrument/parameters/slave' in h5data and
                'entry1/instrument/parameters/master' in h5data):
            master = h5data['entry1/instrument/parameters/master'][0]
            slave = h5data['entry1/instrument/parameters/slave'][0]
        else:
            master = 1
            if abs(chopper2_speed[0]) > 10:
                slave = 2
            elif abs(chopper3_speed[0]) > 10:
                slave = 3
            else:
                slave = 4

        speeds = np.array([chopper1_speed,
                           chopper2_speed,
                           chopper3_speed,
                           chopper4_speed])

        phases = np.array([np.zeros_like(ch2phase),
                           ch2phase,
                           ch3phase,
                           ch4phase])

        return master, slave, speeds[0] / 60., phases[slave - 1]


def basename_datafile(pth):
    """
    Given a NeXUS path return the basename minus the file extension.
    Parameters
    ----------
    pth : str

    Returns
    -------
    basename : str

    Examples
    --------
    >>> basename_datafile('a/b/c.nx.hdf')
    'c'
    """

    basename = os.path.basename(pth)
    return basename.split('.nx.hdf')[0]


def number_datafile(run_number, prefix='PLP'):
    """
    Given a run number figure out what the file name is.
    Given a file name, return the filename with the .nx.hdf extension

    Parameters
    ----------
    run_number : int or str

    prefix : str, optional
        The instrument prefix. Only used if `run_number` is an int

    Returns
    -------
    file_name : str

    Examples
    --------
    >>> number_datafile(708)
    'PLP0000708.nx.hdf'
    >>> number_datafile(708, prefix='QKK')
    'QKK0000708.nx.hdf'
    >>> number_datafile('PLP0000708.nx.hdf')
    'PLP0000708.nx.hdf'
    """
    try:
        num = abs(int(run_number))
        # you got given a run number
        return '{0}{1:07d}.nx.hdf'.format(prefix, num)
    except ValueError:
        # you may have been given full filename
        if run_number.endswith('.nx.hdf'):
            return run_number
        else:
            return run_number + '.nx.hdf'


def datafile_number(fname):
    """
    From a filename figure out what the run number was

    Parameters
    ----------
    fname : str
        The filename to be processed

    Returns
    -------
    run_number : int
        The run number

    Examples
    --------
    >>> datafile_number('PLP0000708.nx.hdf')
    708

    """

    regex = re.compile(".*PLP([0-9]{7}).nx.hdf")
    _fname = os.path.basename(fname)
    r = regex.search(_fname)

    if r:
        return int(r.groups()[0])

    return None


class ReflectNexus(object):
    def __init__(self, h5data, prefix):
        with _possibly_open_hdf_file(h5data, 'r') as f:
            self.cat = Catalogue(f)

        self.processed_spectrum = dict()

        # _arguments is a dict that contains all the parameters used to call
        # `process`. If the arguments don't change then you shouldn't need to
        # call process again, thereby saving time.
        self._arguments = {}

        self.prefix = prefix

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        elif item in self.processed_spectrum:
            return self.processed_spectrum[item]
        else:
            raise AttributeError

    def _short_circuit_process(self, _arguments):
        """
        Returns the truth that two sets of arguments from successive calls to
        the `process` method are the same.

        Parameters
        ----------
        _arguments : dict
            arguments passed to the `process` method

        Returns
        -------
        val : bool
            Truth that __arguments is the same as self.__arguments
        """
        _arguments.pop('self', None)

        return _dict_compare(_arguments, self._arguments)

    def write_spectrum_dat(self, f, scanpoint=0):
        """
        This method writes a dat representation of the corrected spectrum to
        file.

        Parameters
        ----------
        f : file-like or str
            The file to write the spectrum to, or a str that specifies the file
            name
        scanpoint : int
            Which scanpoint to write

        Returns
        -------
        processed : bool
            If the file hasn't been processed then the `processed is False` and
            vice versa
        """
        if self.processed_spectrum is None:
            return False

        m_lambda = self.processed_spectrum['m_lambda'][scanpoint]
        m_spec = self.processed_spectrum['m_spec'][scanpoint]
        m_spec_sd = self.processed_spectrum['m_spec_sd'][scanpoint]
        m_lambda_fwhm = self.processed_spectrum['m_lambda_fwhm'][scanpoint]

        stacked_data = np.c_[m_lambda, m_spec, m_spec_sd, m_lambda_fwhm]
        np.savetxt(f, stacked_data, delimiter='\t')

        return True

    def write_spectrum_xml(self, f, scanpoint=0):
        """
        This method writes an XML representation of the corrected spectrum to
        file.

        Parameters
        ----------
        f : file-like or str
            The file to write the spectrum to, or a str that specifies the file
            name
        scanpoint : int
            Which scanpoint to write
        """
        if self.processed_spectrum is None:
            return

        s = string.Template(spectrum_template)
        d = dict()
        d['title'] = self.cat.sample_name
        d['time'] = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())

        m_lambda = self.processed_spectrum['m_lambda']
        m_spec = self.processed_spectrum['m_spec']
        m_spec_sd = self.processed_spectrum['m_spec_sd']
        m_lambda_fwhm = self.processed_spectrum['m_lambda_fwhm']

        # sort the data
        sorted = np.argsort(self.m_lambda[0])

        r = m_spec[:, sorted]
        lmda = m_lambda[:, sorted]
        dlmda = m_lambda_fwhm[:, sorted]
        dr = m_spec_sd[:, sorted]
        d['n_spectra'] = self.processed_spectrum['n_spectra']
        d['runnumber'] = 'PLP{:07d}'.format(self.cat.datafile_number)

        d['r'] = repr(r[scanpoint].tolist()).strip(',[]')
        d['dr'] = repr(dr[scanpoint].tolist()).strip(',[]')
        d['lmda'] = repr(lmda[scanpoint].tolist()).strip(',[]')
        d['dlmda'] = repr(dlmda[scanpoint].tolist()).strip(',[]')
        thefile = s.safe_substitute(d)

        with possibly_open_file(f, 'wb') as g:
            if 'b' in g.mode:
                thefile = thefile.encode('utf-8')

            g.write(thefile)
            g.truncate()

        return True

    def plot(self, point=0, fig=None):
        """
        Plot a processed spectrum.

        Requires matplotlib be installed.

        Parameters
        ----------
        point: int or sequence, optional
            The spectrum number to be plotted. By default the first spectrum
            will be plotted. Pass `-1` to plot all spectra at once.
        fig: Figure instance, optional
            If `fig` is not supplied then a new figure is created. Otherwise
            the graph is created on the current axes on the supplied figure.

        Returns
        -------
        fig, ax : :class:`matplotlib.Figure`, :class:`matplotlib.Axes`
            `matplotlib` figure and axes objects.

        """
        lam, spec, spec_sd, _ = self.spectrum

        import matplotlib.pyplot as plt
        if fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            ax = fig.gca()

        if hasattr(point, 'len'):
            for p in point:
                ax.plot(lam[p], spec[p])
        elif point == -1:
            for p in range(len(lam)):
                ax.plot(lam[p], spec[p])
        else:
            ax.plot(lam[point], spec[point])

        return fig, ax

    @property
    def spectrum(self):
        return (self.processed_spectrum['m_lambda'],
                self.processed_spectrum['m_spec'],
                self.processed_spectrum['m_spec_sd'],
                self.processed_spectrum['m_lambda_fwhm'])


class PlatypusNexus(ReflectNexus):
    """
    Processes Platypus NeXus files to produce an intensity vs wavelength
    spectrum

    Parameters
    ----------
    h5data : HDF5 NeXus file or str
        An HDF5 NeXus file for Platypus, or a `str` containing the path
        to one
    """

    def __init__(self, h5data):
        """
        Initialises the PlatypusNexus object.
        """
        super(PlatypusNexus, self).__init__(h5data, 'PLP')

    def process(self, h5norm=None, lo_wavelength=2.5, hi_wavelength=19.,
                background=True, direct=False, omega=None, twotheta=None,
                rebin_percent=1., wavelength_bins=None, normalise=True,
                integrate=-1, eventmode=None, event_folder=None,
                peak_pos=None, peak_pos_tol=0.0025,
                background_mask=None, normalise_bins=True,
                manual_beam_find=None, event_filter=None, **kwds):
        r"""
        Processes the ProcessNexus object to produce a time of flight spectrum.
        The processed spectrum is stored in the `processed_spectrum` attribute.
        The specular spectrum is also returned from this function.

        Parameters
        ----------
        h5norm : str or HDF5 NeXus file
            If a str then `h5norm` is a path to the floodfield data, otherwise
            it is a hdf5 file handle containing the floodfield data.
        lo_wavelength : float
            The low wavelength cutoff for the rebinned data (A).
        hi_wavelength : float
            The high wavelength cutoff for the rebinned data (A).
        background : bool
            Should a background subtraction be carried out?
        direct : bool
            Is it a direct beam you measured? This is so a gravity correction
            can be applied.
        omega : float
            Expected angle of incidence of beam. If this is None, then the
            rough angle of incidence is obtained from the NeXus file.
        twotheta : float
            Expected two theta value of specular beam. If this is None then
            the rough angle of incidence is obtained from the NeXus file.
        rebin_percent : float
            Specifies the rebinning percentage for the spectrum.  If
            `rebin_percent is None`, then no rebinning is done.
        wavelength_bins : array_like
            The wavelength bins for rebinning.  If `wavelength_bins is not
            None` then the `rebin_percent` parameter is ignored.
        normalise : bool
            Normalise by the monitor counts.
        integrate : int
            Specifies which scanpoints to use.

             - integrate == -1
               the spectrum is integrated over all the scanpoints.
             - integrate >= 0
               the individual spectra are calculated individually.
               If `eventmode is not None`, or `event_filter is not None` then
               integrate specifies which scanpoint to examine.

        eventmode : None or array_like
            If eventmode is `None` then the integrated detector image is used.
            If eventmode is an array then the array specifies the integration
            times (in seconds) for the detector image, e.g. [0, 20, 30] would
            result in two spectra. The first would contain data for 0 s to 20s,
            the second would contain data for 20 s to 30 s.  This option can
            only be used when `integrate >= -1`.
            If eventmode has zero length (e.g. []), then a single time interval
            for the entire acquisition is used, [0, acquisition_time].  This
            would source the image from the eventmode file, rather than the
            NeXUS file. The two approaches will probably not give
            identical results, because the eventmode method adjusts the total
            acquisition time and beam monitor counts to the frame number of the
            last event detected (which may be quite different if the count rate
            is very low). This parameter is disregarded if `event_filter` is
            provided.
        event_folder : None or str
            Specifies the path for the eventmode data. If
            `event_folder is None` then the eventmode data is assumed to reside
            in the same directory as the NeXUS file. If event_folder is a
            string, then the string specifies the path to the eventmode data.
        peak_pos : -1, None, or (float, float)
            Options for finding specular peak position and peak standard
            deviation.

            - \-1
               use `manual_beam_find`.
            - None
               use the automatic beam finder, falling back to
               `manual_beam_find` if it's provided.
            - (float, float)
               specify the peak and peak standard deviation.

        peak_pos_tol : float
            Convergence tolerance for the beam position and width to be
            accepted from successive beam-finder calculations; see the
            `tol` parameter in the `find_specular_ridge` function.
        background_mask : array_like
            An array of bool that specifies which y-pixels to use for
            background subtraction.  Should be the same length as the number of
            y pixels in the detector image.  Otherwise an automatic mask is
            applied (if background is True).
        normalise_bins : bool
            Divides the intensity in each wavelength bin by the width of the
            bin. This allows one to compare spectra even if they were processed
            with different rebin percentages.
        manual_beam_find : callable, optional
            A function which allows the location of the specular ridge to be
            determined. Has the signature `f(detector, detector_err)` where
            `detector` and `detector_err` is the detector image and its
            uncertainty. `detector` and `detector_err` have shape (n, t, y)
            where `n` is the number of detector images, `t` is the number of
            time of flight bins and `y` is the number of y pixels. The function
            should return a tuple,
            `(centre, centre_sd, lopx, hipx, background_pixels)`. `centre`,
            `centre_sd`, `lopx`, `hipx` should be arrays of shape `(n, )`,
            specifying the beam centre, beam width (standard deviation), lowest
            pixel of foreground region, highest pixel of foreground region.
            `background_pixels` is a list of length `n`. Each of the entries
            should contain arrays of pixel numbers that specify the background
            region for each of the detector images.
        event_filter : callable, optional
            A function, that processes the event stream, returning a `detector`
            array, and a `frame_count` array. `detector` has shape
            `(N, T, Y, X)`, where `N` is the number of detector images, `T` is
            the number of time bins (`len(t_bins)`), etc. `frame_count` has
            shape `(N,)` and contains the number of frames for each of the
            detector images. The frame_count is used to determine what fraction
            of the overall monitor counts should be ascribed to each detector
            image (by dividing by the total number of frames). The function has
            signature:

            detector, frame_count = event_filter(loaded_events,
                                                 t_bins=None,
                                                 y_bins=None,
                                                 x_bins=None)

            `loaded_events` is a 4-tuple of numpy arrays:
            `(f_events, t_events, y_events, x_events)`, where `f_events`
            contains the frame number for each neutron, landing at position
            `x_events, y_events` on the detector, with time-of-flight
            `t_events`.

        Notes
        -----
        After processing this object contains the following the following
        attributes:

        - path - path to the data file
        - datafilename - name of the datafile
        - datafile_number - datafile number.
        - m_topandtail - the corrected 2D detector image, (n_spectra, TOF, Y)
        - m_topandtail_sd - corresponding standard deviations
        - n_spectra - number of spectra in processed data
        - bm1_counts - beam montor counts, (n_spectra,)
        - m_spec - specular intensity, (n_spectra, TOF)
        - m_spec_sd - corresponding standard deviations
        - m_beampos - beam_centre for each spectrum, (n_spectra, )
        - m_lambda - wavelengths for each spectrum, (n_spectra, TOF)
        - m_lambda_fwhm - corresponding FWHM of wavelength distribution
        - m_lambda_hist - wavelength bins for each spectrum, (n_spectra, TOF)
        - m_spec_tof - TOF for each wavelength bin, (n_spectra, TOF)
        - mode - the Platypus mode, e.g. FOC/MT/POL/POLANAL/SB/DB
        - detector_z - detector height, (n_spectra, )
        - detector_y - sample-detector distance, (n_spectra, )
        - domega - collimation uncertainty
        - lopx - lowest extent of specular beam (in y pixels), (n_spectra, )
        - hipx - highest extent of specular beam (in y pixels), (n_spectra, )

        Returns
        -------
        m_lambda, m_spec, m_spec_sd: np.ndarray
            Arrays containing the wavelength, specular intensity as a function
            of wavelength, standard deviation of specular intensity

        """

        # it can be advantageous to save processing time if the arguments
        # haven't changed
        _arguments = locals().copy()
        _arguments.pop('_arguments', None)
        _arguments.pop('self', None)
        # if you've already processed, then you may not need to process again
        if (self.processed_spectrum and
                self._short_circuit_process(_arguments)):
            return (self.processed_spectrum['m_lambda'],
                    self.processed_spectrum['m_spec'],
                    self.processed_spectrum['m_spec_sd'])
        else:
            self._arguments = _arguments

        cat = self.cat

        scanpoint = 0

        # beam monitor counts for normalising data
        bm1_counts = cat.bm1_counts.astype('float64')

        # TOF bins
        TOF = cat.t_bins.astype('float64')

        # This section controls how multiple detector images are handled.
        # We want event streaming.
        if eventmode is not None or event_filter is not None:
            scanpoint = integrate
            if integrate == -1:
                scanpoint = 0

            output = self.process_event_stream(scanpoint=scanpoint,
                                               frame_bins=eventmode,
                                               event_folder=event_folder,
                                               event_filter=event_filter)
            detector, frame_count, bm1_counts = output

            start_time = np.zeros(np.size(detector, 0))
            if cat.start_time is not None:
                start_time += cat.start_time[scanpoint]
                start_time[1:] += (np.cumsum(frame_count)[:-1] /
                                   cat.frequency[scanpoint])
        else:
            # we don't want detector streaming
            detector = cat.detector
            scanpoint = 0

            # integrate over all spectra
            if integrate == -1:
                detector = np.sum(detector, 0)[np.newaxis, ]
                bm1_counts[:] = np.sum(bm1_counts)

            start_time = np.zeros(np.size(detector, 0))
            if cat.start_time is not None:
                for idx in range(start_time.size):
                    start_time[idx] = cat.start_time[idx]

        n_spectra = np.size(detector, 0)

        # Up until this point detector.shape=(N, T, Y)
        # pre-average over x, leaving (n, t, y) also convert to dp
        detector = np.sum(detector, axis=3, dtype='float64')

        # detector shape should now be (n, t, y)
        # calculate the counting uncertainties
        detector_sd = np.sqrt(detector)
        bm1_counts_sd = np.sqrt(bm1_counts)

        # detector normalisation with a water file
        if h5norm is not None:
            x_bins = cat.x_bins

            with _possibly_open_hdf_file(h5norm, 'r') as f:
                # shape (y,)
                detector_norm, detector_norm_sd = create_detector_norm(
                    f, x_bins[0], x_bins[1])

            # detector has shape (N, T, Y), shape of detector_norm should
            # broadcast to (1, 1, y)
            # TODO: Correlated Uncertainties?
            detector, detector_sd = EP.EPdiv(detector, detector_sd,
                                             detector_norm, detector_norm_sd)

        # shape of these is (n_spectra, TOFbins)
        m_spec_tof_hist = np.zeros((n_spectra, np.size(TOF, 0)),
                                   dtype='float64')
        m_lambda_hist = np.zeros((n_spectra, np.size(TOF, 0)),
                                 dtype='float64')
        m_spec_tof_hist[:] = TOF

        """
        chopper to detector distances
        note that if eventmode is specified the n_spectra is NOT
        equal to the number of entries in e.g. /longitudinal_translation
        this means you have to copy values in from the correct scanpoint
        """
        flight_distance = np.zeros(n_spectra, dtype='float64')
        d_cx = np.zeros(n_spectra, dtype='float64')
        detpositions = np.zeros(n_spectra, dtype='float64')

        # The angular divergence of the instrument
        domega = np.zeros(n_spectra, dtype='float64')
        estimated_beam_width = np.zeros(n_spectra, dtype='float64')

        phase_angle = np.zeros(n_spectra, dtype='float64')

        # process each of the spectra taken in the detector image
        original_scanpoint = scanpoint
        for idx in range(n_spectra):
            freq = cat.frequency[scanpoint]

            # calculate the angular divergence
            domega[idx] = general.div(cat.ss2vg[scanpoint],
                                      cat.ss3vg[scanpoint],
                                      (cat.slit3_distance[0] -
                                       cat.slit2_distance[0]))[0]

            """
            estimated beam width in pixels at detector
            """
            L23 = cat.slit3_distance[idx] - cat.slit2_distance[0]
            L3det = (cat.dy[idx] +
                     cat.sample_distance[0] - cat.slit3_distance[idx])
            ebw = general.height_of_beam_after_dx(cat.ss2vg[scanpoint],
                                                  cat.ss3vg[scanpoint],
                                                  L23,
                                                  L3det)
            umb, penumb = ebw
            # convolve in detector resolution (~2.2 mm?)
            # first convert to beam sd, convolve in detector, and expand sd
            # back to total foreground width
            # use average of umb and penumb, the calc assumes a rectangular
            # distribution
            penumb = (np.sqrt((0.289 * 0.5 * (umb + penumb))**2. + 2.2**2) *
                      EXTENT_MULT * 2)
            # we need it in pixels
            estimated_beam_width[idx] = penumb / Y_PIXEL_SPACING

            """
            work out the total flight length
            IMPORTANT: this varies as a function of twotheta. This is
            because the Platypus detector does not move on an arc.
            At high angles chod can be ~ 0.75% different. This is will
            visibly shift fringes.
            """
            if omega is None:
                omega = cat.omega[scanpoint]
            if twotheta is None:
                twotheta = cat.twotheta[scanpoint]
            output = self.chod(omega, twotheta, scanpoint=scanpoint)
            flight_distance[idx], d_cx[idx] = output

            # calculate nominal phase openings
            output = self.phase_angle(scanpoint)
            phase_angle[scanpoint], master_opening = output

            """
            toffset - the time difference between the magnet pickup on the
            choppers (TTL pulse), which is situated in the middle of the
            chopper window, and the trailing edge of chopper 1, which is
            supposed to be time0.  However, if there is a phase opening this
            time offset has to be relocated slightly, as time0 is not at the
            trailing edge.
            """
            t_offset = self._platypus_time_offset(cat.chopper1_phase_offset[0],
                                                  master_opening,
                                                  freq,
                                                  phase_angle[scanpoint],
                                                  d_cx[0],
                                                  flight_distance[idx],
                                                  m_spec_tof_hist[idx])

            m_spec_tof_hist[idx] -= t_offset

            detpositions[idx] = cat.dy[scanpoint]

            if eventmode is not None or event_filter is not None:
                m_spec_tof_hist[:] = TOF - t_offset
                flight_distance[:] = flight_distance[0]
                detpositions[:] = detpositions[0]
                domega[:] = domega[0]
                estimated_beam_width[:] = estimated_beam_width[0]
                d_cx[:] = d_cx[0]
                phase_angle[:] = phase_angle[0]
                break
            else:
                scanpoint += 1

        scanpoint = original_scanpoint

        # convert TOF to lambda
        # m_spec_tof_hist (n, t) and chod is (n,)
        m_lambda_hist = general.velocity_wavelength(
            1.e3 * flight_distance[:, np.newaxis] / m_spec_tof_hist)

        m_lambda = 0.5 * (m_lambda_hist[:, 1:] + m_lambda_hist[:, :-1])
        TOF -= t_offset

        # gravity correction if direct beam
        if direct:
            # TODO: Correlated Uncertainties?
            output = correct_for_gravity(detector,
                                         detector_sd,
                                         m_lambda,
                                         self.cat.collimation_distance,
                                         self.cat.dy,
                                         lo_wavelength,
                                         hi_wavelength)
            detector, detector_sd, m_gravcorrcoefs = output

        # where is the specular ridge?
        if peak_pos == -1:
            # you always want to find the beam manually
            ret = manual_beam_find(detector, detector_sd)
            beam_centre, beam_sd, lopx, hipx, bp = ret

            full_backgnd_mask = np.zeros_like(detector, dtype=bool)
            for i, v in enumerate(bp):
                full_backgnd_mask[i, :, v] = True

        elif peak_pos is None:
            # use the auto finder, falling back to manual_beam_find
            ret = find_specular_ridge(detector,
                                      detector_sd,
                                      tol=peak_pos_tol,
                                      manual_beam_find=manual_beam_find)
            beam_centre, beam_sd, lopx, hipx, full_backgnd_mask = ret
        else:
            # the specular ridge has been specified
            beam_centre = np.ones(n_spectra) * peak_pos[0]
            beam_sd = np.ones(n_spectra) * peak_pos[1]
            lopx, hipx, bp = fore_back_region(beam_centre,
                                              beam_sd)

            full_backgnd_mask = np.zeros_like(detector, dtype=bool)
            for i, v in enumerate(bp):
                full_backgnd_mask[i, :, v] = True

        lopx = lopx.astype(int)
        hipx = hipx.astype(int)

        # Warning if the beam appears to be much broader than the divergence
        # would predict. The use of 30% tolerance is a guess. This might happen
        # if the beam finder includes incoherent background region by mistake.
        if not np.allclose(estimated_beam_width, hipx - lopx + 1, rtol=0.3):
            warnings.warn("The foreground width (%s) estimate"
                          " does not match the divergence of the beam (%s)."
                          " Consider checking with manual beam finder." %
                          (str(hipx - lopx + 1),
                           str(estimated_beam_width)))

        if np.size(beam_centre) != n_spectra:
            raise RuntimeError('The number of beam centres should be equal'
                               ' to the number of detector images.')

        '''
        Rebinning in lambda for all detector
        Rebinning is the default option, but sometimes you don't want to.
        detector shape input is (n, t, y)
        '''
        if wavelength_bins is not None:
            rebinning = wavelength_bins
        elif 0. < rebin_percent < 15.:
            rebinning = calculate_wavelength_bins(lo_wavelength,
                                                  hi_wavelength,
                                                  rebin_percent)

        # rebin_percent percentage is zero. No rebinning, just cutoff
        # wavelength
        else:
            rebinning = m_lambda_hist[0, :]
            rebinning = rebinning[np.searchsorted(rebinning, lo_wavelength):
                                  np.searchsorted(rebinning, hi_wavelength)]

        """
        now do the rebinning for all the N detector images
        rebin.rebinND could do all of these at once.  However, m_lambda_hist
        could vary across the range of spectra.  If it was the same I could
        eliminate the loop.
        """
        output = []
        output_sd = []
        for idx in range(n_spectra):
            # TODO: Correlated Uncertainties?
            plane, plane_sd = rebin_along_axis(detector[idx],
                                               m_lambda_hist[idx],
                                               rebinning,
                                               y1_sd=detector_sd[idx])
            output.append(plane)
            output_sd.append(plane_sd)

        detector = np.array(output)
        detector_sd = np.array(output_sd)

        if len(detector.shape) == 2:
            detector = detector[np.newaxis, ]
            detector_sd = detector_sd[np.newaxis, ]

        # (1, T)
        m_lambda_hist = np.atleast_2d(rebinning)

        """
        Divide the detector intensities by the width of the wavelength bin.
        This is so the intensities between different rebinning strategies can
        be compared.
        """
        if normalise_bins:
            div = 1 / np.diff(m_lambda_hist[0])[:, np.newaxis]
            detector, detector_sd = EP.EPmulk(detector,
                                              detector_sd,
                                              div)

        # convert the wavelength base to a timebase
        m_spec_tof_hist = (0.001 * flight_distance[:, np.newaxis] /
                           general.wavelength_velocity(m_lambda_hist))

        m_lambda = 0.5 * (m_lambda_hist[:, 1:] + m_lambda_hist[:, :-1])

        m_spec_tof = (0.001 * flight_distance[:, np.newaxis] /
                      general.wavelength_velocity(m_lambda))

        m_spec = np.zeros((n_spectra, np.size(detector, 1)))
        m_spec_sd = np.zeros_like(m_spec)

        # background subtraction
        if background:
            if background_mask is not None:
                # background_mask is (Y), need to make 3 dimensional (N, T, Y)
                # first make into (T, Y)
                backgnd_mask = np.repeat(background_mask[np.newaxis, :],
                                         detector.shape[1],
                                         axis=0)
                # make into (N, T, Y)
                full_backgnd_mask = np.repeat(backgnd_mask[np.newaxis, :],
                                              n_spectra,
                                              axis=0)

            # TODO: Correlated Uncertainties?
            detector, detector_sd = background_subtract(detector,
                                                        detector_sd,
                                                        full_backgnd_mask)

        '''
        top and tail the specular beam with the known beam centres.
        All this does is produce a specular intensity with shape (N, T),
        i.e. integrate over specular beam
        '''
        for i in range(n_spectra):
            m_spec[i] = np.sum(detector[i, :, lopx[i]: hipx[i] + 1], axis=1)
            sd = np.sum(detector_sd[i, :, lopx[i]: hipx[i] + 1] ** 2,
                        axis=1)
            m_spec_sd[i] = np.sqrt(sd)

        # assert np.isfinite(m_spec).all()
        # assert np.isfinite(m_specSD).all()
        # assert np.isfinite(detector).all()
        # assert np.isfinite(detectorSD).all()

        # normalise by beam monitor 1.
        if normalise:
            m_spec, m_spec_sd = EP.EPdiv(m_spec,
                                         m_spec_sd,
                                         bm1_counts[:, np.newaxis],
                                         bm1_counts_sd[:, np.newaxis])

            output = EP.EPdiv(detector,
                              detector_sd,
                              bm1_counts[:, np.newaxis, np.newaxis],
                              bm1_counts_sd[:, np.newaxis, np.newaxis])
            detector, detector_sd = output

        '''
        now work out dlambda/lambda, the resolution contribution from
        wavelength.
        van Well, Physica B,  357(2005) pp204-207), eqn 4.
        this is only an approximation for our instrument, as the 2nd and 3rd
        discs have smaller openings compared to the master chopper.
        Therefore the burst time needs to be looked at.
        '''
        tau_da = m_spec_tof_hist[:, 1:] - m_spec_tof_hist[:, :-1]

        m_lambda_fwhm = (
            resolution_double_chopper(m_lambda,
                                      z0=d_cx[:, np.newaxis] / 1000.,
                                      freq=cat.frequency[:, np.newaxis],
                                      L=flight_distance[:, np.newaxis] / 1000.,
                                      H=cat.ss2vg[original_scanpoint] / 1000.,
                                      xsi=phase_angle[:, np.newaxis],
                                      tau_da=tau_da))

        m_lambda_fwhm *= m_lambda

        # put the detector positions and mode into the dictionary as well.
        detector_z = cat.dz
        detector_y = cat.dy
        mode = cat.mode

        d = dict()
        d['path'] = cat.path
        d['datafilename'] = cat.filename
        d['datafile_number'] = cat.datafile_number

        if h5norm is not None:
            if type(h5norm) == h5py.File:
                d['normfilename'] = h5norm.filename
            else:
                d['normfilename'] = h5norm
        d['m_topandtail'] = detector
        d['m_topandtail_sd'] = detector_sd
        d['n_spectra'] = n_spectra
        d['bm1_counts'] = bm1_counts
        d['m_spec'] = m_spec
        d['m_spec_sd'] = m_spec_sd
        d['m_beampos'] = beam_centre
        d['m_beampos_sd'] = beam_sd
        d['m_lambda'] = m_lambda
        d['m_lambda_fwhm'] = m_lambda_fwhm
        d['m_lambda_hist'] = m_lambda_hist
        d['m_spec_tof'] = m_spec_tof
        d['mode'] = mode
        d['detector_z'] = detector_z
        d['detector_y'] = detector_y
        d['domega'] = domega
        d['lopx'] = lopx
        d['hipx'] = hipx
        d['start_time'] = start_time

        self.processed_spectrum = d
        return m_lambda, m_spec, m_spec_sd

    def _platypus_time_offset(self, master_phase_offset, master_opening,
                              freq, phase_angle, z0, flight_distance,
                              tof_hist):
        """
        Timing offsets for Platypus chopper system, includes a gravity
        correction for phase angle
        """
        # calculate initial time offset
        p_offset = 1.e6 * master_phase_offset / (2. * 360. * freq)
        t_offset = (p_offset +
                    1.e6 * master_opening / 2 / (2 * np.pi) / freq -
                    1.e6 * phase_angle / (360 * 2 * freq))

        ###########################################
        # now make a gravity correction to t_offset
        # work out velocities for each bin edge
        velocities = 1.e3 * flight_distance / (tof_hist - t_offset)

        angles = find_trajectory(self.cat.collimation_distance / 1000.,
                                 0,
                                 velocities)

        # work out distance from 1st coll slit to middle of chopper pair
        # TODO ASSUMES CHOPPER 1 IS MASTER, FIX SO IT COULD BE ANY MASTER
        d_c1 = -self.cat.slit2_distance
        d_slave = d_c1 + z0

        corr_t_offset = np.zeros_like(tof_hist)
        corr_t_offset += master_opening / 2 / (2 * np.pi) / freq

        for i, (velocity, angle) in enumerate(zip(velocities, angles)):
            parab = parabola(angle, velocity)
            h_c1 = parab(d_c1 / 1000.) * 1000.
            h_slave = parab(d_slave / 1000.) * 1000.
            # angle_corr comes about because the parabolic nature of long
            # wavelength neutrons creates an apparent phase opening
            angle_corr = np.degrees(np.arctan((h_slave - h_c1) /
                                              DISCRADIUS))
            # master_corr comes about because the beam for long wavelength
            # neutrons is lower than the optical axis, so it takes a little
            # longer for the beam to start travelling through chopper system.
            # as such you need to decrease their tof by increasing the
            # t_offset
            master_corr = -np.degrees(np.arctan(h_c1 / DISCRADIUS))
            corr_t_offset[i] += ((master_phase_offset + master_corr) /
                                 (2. * 360. * freq))
            corr_t_offset[i] -= ((phase_angle + angle_corr) /
                                 (360 * 2 * freq))
        corr_t_offset *= 1e6

        return corr_t_offset

    def phase_angle(self, scanpoint=0):
        """
        Calculates the phase angle for a given scanpoint

        Parameters
        ----------
        scanpoint : int
            The scanpoint you're interested in

        Returns
        -------
        phase_angle, master_opening : float
            The phase angle in degrees, and the angular opening of the master
            chopper
        """
        cat = self.cat
        master = cat.master
        slave = cat.slave
        disc_phase = cat.phase[scanpoint]
        phase_angle = 0

        if master == 1:
            phase_angle += 0.5 * O_C1d
            master_opening = O_C1
        elif master == 2:
            phase_angle += 0.5 * O_C2d
            master_opening = O_C2
        elif master == 3:
            phase_angle += 0.5 * O_C3d
            master_opening = O_C3

        if slave == 2:
            phase_angle += 0.5 * O_C2d
            phase_angle += -disc_phase - cat.chopper2_phase_offset[0]
        elif slave == 3:
            phase_angle += 0.5 * O_C3d
            phase_angle += -disc_phase - cat.chopper3_phase_offset[0]
        elif slave == 4:
            phase_angle += 0.5 * O_C4d
            phase_angle += disc_phase - cat.chopper4_phase_offset[0]

        return phase_angle, master_opening

    def chod(self, omega=0., twotheta=0., scanpoint=0):
        """
        Calculates the flight length of the neutrons in the Platypus
        instrument.

        Parameters
        ----------
        omega : float, optional
            Rough angle of incidence
        twotheta : float, optional
            Rough 2 theta angle
        scanpoint : int, optional
            Which dataset is being considered

        Returns
        -------
        chod, d_cx : float, float
            Flight distance (mm), distance between chopper discs (mm)
        """

        chod = 0

        # guide 1 is the single deflection mirror (SB)
        # its distance is from chopper 1 to the middle of the mirror (1m long)
        # guide 2 is the double deflection mirror (DB)
        # its distance is from chopper 1 to the middle of the second of the
        # compound mirrors! (a bit weird, I know).

        cat = self.cat
        mode = cat.mode

        # Find out chopper pairing
        master = cat.master
        slave = cat.slave

        d_cx = 0

        if master == 1:
            chod = 0
        elif master == 2:
            chod -= cat.chopper2_distance[0]
            d_cx -= chod
        elif master == 3:
            chod -= cat.chopper3_distance[0]
            d_cx -= chod
        else:
            raise ValueError("Chopper pairing should be one of '12', '13',"
                             "'14', '23', '24', '34'")

        if slave == 2:
            chod -= cat.chopper2_distance[0]
            d_cx += cat.chopper2_distance[0]
        elif slave == 3:
            chod -= cat.chopper3_distance[0]
            d_cx += cat.chopper3_distance[0]
        elif slave == 4:
            chod -= cat.chopper4_distance[0]
            d_cx += cat.chopper4_distance[0]

        # start of flight length is midway between master and slave, but master
        # may not necessarily be disk 1. However, all instrument lengths are
        # measured from disk1
        chod /= 2.

        if mode in ['FOC', 'POL', 'MT', 'POLANAL']:
            chod += cat.sample_distance[0]
            chod += cat.dy[scanpoint] / np.cos(np.radians(twotheta))

        elif mode == 'SB':
            # assumes guide1_distance is in the MIDDLE OF THE MIRROR
            chod += cat.guide1_distance[0]
            chod += ((cat.sample_distance[0] - cat.guide1_distance[0]) /
                     np.cos(np.radians(omega)))
            if twotheta > omega:
                chod += (cat.dy[scanpoint] /
                         np.cos(np.radians(twotheta - omega)))
            else:
                chod += (cat.dy[scanpoint] /
                         np.cos(np.radians(omega - twotheta)))

        elif mode == 'DB':
            # guide2_distance in in the middle of the 2nd compound mirror
            # guide2_distance - longitudinal length from midpoint1 -> midpoint2
            #  + direct length from midpoint1->midpoint2
            chod += (cat.guide2_distance[0] +
                     600. * np.cos(np.radians(1.2)) *
                     (1 - np.cos(np.radians(2.4))))

            # add on distance from midpoint2 to sample
            chod += ((cat.sample_distance[0] - cat.guide2_distance[0]) /
                     np.cos(np.radians(4.8)))

            # add on sample -> detector
            if twotheta > omega:
                chod += (cat.dy[scanpoint] /
                         np.cos(np.radians(twotheta - 4.8)))
            else:
                chod += (cat.dy[scanpoint] /
                         np.cos(np.radians(4.8 - twotheta)))

        return chod, d_cx

    def process_event_stream(self, t_bins=None, x_bins=None, y_bins=None,
                             frame_bins=None, scanpoint=0, event_folder=None,
                             event_filter=None):
        """
        Processes the event mode dataset for the NeXUS file. Assumes that
        there is a event mode directory in the same directory as the NeXUS
        file, as specified by in 'entry1/instrument/detector/daq_dirname'

        Parameters
        ----------
        t_bins : array_like, optional
            specifies the time bins required in the image
        x_bins : array_like, optional
            specifies the x bins required in the image
        y_bins : array_like, optional
            specifies the y bins required in the image
        scanpoint : int, optional
            Scanpoint you are interested in
        event_folder : None or str
            Specifies the path for the eventmode data. If
            `event_folder is None` then the eventmode data is assumed to reside
            in the same directory as the NeXUS file. If event_folder is a
            string, then the string specifies the path to the eventmode data.
        frame_bins : array_like, optional
            specifies the frame bins required in the image. If
            frame_bins = [5, 10, 120] you will get 2 images.  The first starts
            at 5s and finishes at 10s. The second starts at 10s and finishes
            at 120s. If frame_bins has zero length, e.g. [], then a single
            interval consisting of the entire acquisition time is used:
            [0, acquisition_time]. If `event_filter` is provided then this
            parameter is ignored.
        event_filter : callable, optional
            A function, that processes the event stream, returning a `detector`
            array, and a `frame_count` array. `detector` has shape
            `(N, T, Y, X)`, where `N` is the number of detector images, `T` is
            the number of time bins (`len(t_bins)`), etc. `frame_count` has
            shape `(N,)` and contains the number of frames for each of the
            detector images. The frame_count is used to determine what fraction
            of the overall monitor counts should be ascribed to each detector
            image. The function has signature:

            detector, frame_count = event_filter(loaded_events,
                                                 t_bins=None,
                                                 y_bins=None,
                                                 x_bins=None)

            `loaded_events` is a 4-tuple of numpy arrays:
            `(f_events, t_events, y_events, x_events)`, where `f_events`
            contains the frame number for each neutron, landing at position
            `x_events, y_events` on the detector, with time-of-flight
            `t_events`.

        Returns
        -------
        detector, frame_count, bm1_counts : np.ndarray, np.ndarray, np.ndarray

        Create a new detector image based on the t_bins, x_bins, y_bins and
        frame_bins you supply to the method (these should all be lists/numpy
        arrays specifying the edges of the required bins). If these are not
        specified, then the default bins are taken from the nexus file. This
        would essentially return the same detector image as the nexus file.
        However, you can specify the frame_bins list to generate detector
        images based on subdivided periods of the total acquisition.
        For example if frame_bins = [5, 10, 120] you will get 2 images.  The
        first starts at 5s and finishes at 10s. The second starts at 10s
        and finishes at 120s. The frame_bins are clipped to the total
        acquisition time if necessary.
        `frame_count` is how many frames went into making each detector image.

        """
        cat = self.cat

        if not t_bins:
            t_bins = cat.t_bins
        if not y_bins:
            y_bins = cat.y_bins
        if not x_bins:
            x_bins = cat.x_bins
        if frame_bins is None or np.size(frame_bins) == 0:
            frame_bins = [0, cat.time[scanpoint]]

        total_acquisition_time = cat.time[scanpoint]
        frequency = cat.frequency[scanpoint]

        bm1_counts_for_scanpoint = cat.bm1_counts[scanpoint]

        event_directory_name = cat.daq_dirname

        _eventpath = cat.path
        if event_folder is not None:
            _eventpath = event_folder

        stream_filename = os.path.join(_eventpath,
                                       event_directory_name,
                                       'DATASET_%d' % scanpoint,
                                       'EOS.bin')

        with io.open(stream_filename, 'rb') as f:
            last_frame = int(frame_bins[-1] * frequency)
            loaded_events, end_of_last_event = events(f,
                                                      max_frames=last_frame)

        # convert frame_bins to list of filter frames
        frames = framebins_to_frames(np.asfarray(frame_bins) *
                                     frequency)

        if event_filter is not None:
            output = event_filter(loaded_events, t_bins, y_bins, x_bins)
        else:
            output = process_event_stream(loaded_events,
                                          frames,
                                          t_bins,
                                          y_bins,
                                          x_bins)

        detector, frame_count = output

        bm1_counts = (frame_count * bm1_counts_for_scanpoint /
                      total_acquisition_time /
                      frequency)

        return detector, frame_count, bm1_counts


def create_detector_norm(h5norm, x_min, x_max):
    """
    Produces a detector normalisation array for a neutron detector.
    Here we average over N, T and X to provide a relative efficiency for each
    y wire.

    Parameters
    ----------
    h5norm : hdf5 file
        Containing a flood field run (water)
    x_min : float, int
        Minimum x location to use
    x_max : float, int
        Maximum x location to use

    Returns
    -------
    norm, norm_sd : array_like
        1D array containing the normalisation data for each y pixel
    """
    # sum over N and T
    detector = h5norm['entry1/data/hmm']
    x_bin = h5norm['entry1/data/x_bin']

    # find out what pixels to use
    x_low = np.searchsorted(x_bin, x_min, sorter=np.argsort(x_bin))
    x_low = np.argsort(x_bin)[x_low]
    x_high = np.searchsorted(x_bin, x_max, sorter=np.argsort(x_bin))
    x_high = np.argsort(x_bin)[x_high]

    if(x_low > x_high):
        x_low, x_high = x_high, x_low

    norm = np.sum(detector, axis=(0, 1),
                  dtype='float64')

    # By this point you have norm[y][x]
    norm = norm[:, x_low: x_high]
    norm = np.sum(norm, axis=1)

    mean = np.mean(norm)

    return norm / mean, np.sqrt(norm) / mean


def background_subtract(detector, detector_sd, background_mask):
    """
    Background subtraction of Platypus detector image.
    Shape of detector is (N, T, Y), do a linear background subn for each
    (N, T) slice.

    Parameters
    ----------
    detector : np.ndarray
        detector array with shape (N, T, Y).
    detector_sd : np.ndarray
        standard deviations for detector array
    background_mask : array_like
        array of bool with shape (N, T, Y) that specifies which Y pixels to use
        for background subtraction.

    Returns
    -------
    detector, detector_sd : np.ndarray, np.ndarray
        Detector image with background subtracted
    """
    ret = np.zeros_like(detector)
    ret_sd = np.zeros_like(detector)

    for idx in np.ndindex(detector.shape[0: 2]):
        ret[idx], ret_sd[idx] = background_subtract_line(detector[idx],
                                                         detector_sd[idx],
                                                         background_mask[idx])
    return ret, ret_sd


def background_subtract_line(profile, profile_sd, background_mask):
    """
    Performs a linear background subtraction on a 1D peak profile

    Parameters
    ----------
    profile : np.ndarray
        1D profile
    profile_sd : np.ndarray
        standard deviations for profile
    background_mask : array_like
        array of bool that specifies which Y pixels to use for background
        subtraction.

    Returns
    -------
    profile_subt, profile_subt_err : np.ndarray, np.ndarray
        Background subtracted profile and its uncertainty
    """

    # which values to use as a background region
    mask = np.array(background_mask).astype('bool')
    x_vals = np.where(mask)[0]

    if np.size(x_vals) < 2:
        # can't do a background subtraction if you have less than 2 points in
        # the background
        return profile, profile_sd

    try:
        y_vals = profile[x_vals]
    except IndexError:
        print(x_vals)

    y_sdvals = profile_sd[x_vals]
    x_vals = x_vals.astype('float')

    # some SD values may have 0 SD, which will screw up curvefitting.
    y_sdvals = np.where(y_sdvals == 0, 1, y_sdvals)

    # equation for a straight line
    def f(x, a, b):
        return a + b * x

    # estimate the linear fit
    y_bar = np.mean(y_vals)
    x_bar = np.mean(x_vals)
    bhat = np.sum((x_vals - x_bar) * (y_vals - y_bar))
    bhat /= np.sum((x_vals - x_bar) ** 2)
    ahat = y_bar - bhat * x_bar

    # get the weighted fit values
    # we know the absolute sigma values
    popt, pcov = curve_fit(f, x_vals, y_vals, sigma=y_sdvals,
                           p0=np.array([ahat, bhat]), absolute_sigma=True)

    def CI(xx, pcovmat):
        return (pcovmat[0, 0] + pcovmat[1, 0] * xx +
                pcovmat[0, 1] * xx + pcovmat[1, 1] * (xx ** 2))

    bkgd = f(np.arange(np.size(profile, 0)), popt[0], popt[1])

    # now work out confidence intervals
    # TODO, should this be confidence interval or prediction interval?
    # if you try to do a fit which has a singular matrix
    if np.isfinite(pcov).all():
        bkgd_sd = np.asarray([CI(x, pcov) for x in np.arange(len(profile))],
                             dtype='float64')
    else:
        bkgd_sd = np.zeros_like(bkgd)

    bkgd_sd = np.sqrt(bkgd_sd)

    # get the t value for a two sided student t test at the 68.3 confidence
    # level
    bkgd_sd *= t.isf(0.1585, np.size(x_vals, 0) - 2)

    return EP.EPsub(profile, profile_sd, bkgd, bkgd_sd)


def find_specular_ridge(detector, detector_sd, search_increment=50,
                        tol=0.002, manual_beam_find=None):
    """
    Find the specular ridges in a detector(n, t, y) plot.

    Parameters
    ----------
    detector : array_like
        detector array
    detector_sd : array_like
        standard deviations of detector array
    search_increment : int
        specifies the search increment for the location process.
    tol : float
        specifies threshold of fractional change for beam centre to be found
    manual_beam_find : callable, optional
        A function which allows the location of the specular ridge to be
        determined, IFF the autodetection of specular ridge fails.
        Has the signature `f(detector, detector_sd)`. `detector` and
        `detector_sd` have shape (n, t, y) where `n` is the number of detector
        images, `t` is the number of time of flight bins and `y` is the number
        of y pixels. The function should return a tuple,
        `(centre, centre_sd, lopx, hipx, background_pixels)`. `centre`,
        `centre_sd`, `lopx`, `hipx` should be arrays of shape `(n, )`,
        specifying the beam centre, beam width (standard deviation), lowest
        pixel of foreground region, highest pixel of foreground region.
        `background_pixels` is a list of length `n`. Each of the entries
        should contain arrays of pixel numbers that specify the background
        region for each of the detector images.

    Returns
    -------
    centre, SD, lopx, hipx, background_mask : np.ndarrays
        peak centre, standard deviation of peak width, lowest pixel to be
        included from background region, highest pixel to be included from
        background region, array specifying points to be used for background
        subtraction
        `np.size(centre) == n`.

    Notes
    -----
    The search for the beam centre proceeds by taking the last
    `search_increment` time bins, summing over the time axis and finding
    the beam centre and width along the y-axis. It then repeats the process
    with the last `2 * search_increment` time bins. This process is repeated
    until the relative change in beam centre and width is lower than `tol`.
    This process is designed to locate the specular ridge, even in the
    presence of incoherent scattering.
    """
    beam_centre = np.zeros(np.size(detector, 0))
    beam_sd = np.zeros_like(beam_centre)

    # lopx and hipx specify the foreground region to integrate over
    lopx = np.zeros_like(beam_centre, dtype=int)
    hipx = np.zeros_like(beam_centre, dtype=int)

    # background mask specifies which pixels are background
    background_mask = np.zeros_like(detector, dtype=bool)

    search_increment = abs(search_increment)

    n_increments = ((np.size(detector, 1) - search_increment) //
                    search_increment)

    # we want to integrate over the following pixel region
    for j in range(np.size(detector, 0)):
        last_centre = -1.
        last_sd = -1.
        converged = False
        for i in range(n_increments):
            how_many = -search_increment * (1 + i)
            det_subset = detector[j, how_many:]
            det_sd_subset = detector_sd[j, how_many:]

            # Uncertainties code takes a while to run
            # total_y = np.sum(det_subset, axis=0)
            y_cross = np.sum(det_subset, axis=0)
            y_cross_sd = np.sqrt(np.sum(det_sd_subset ** 2., axis=0))

            # find the centroid and gauss peak in the last sections of the TOF
            # plot
            try:
                centroid, gauss_peak = peak_finder(y_cross,
                                                   sigma=y_cross_sd)
            except RuntimeError:
                continue

            if (abs((gauss_peak[0] - last_centre) / last_centre) < tol and
                    abs((gauss_peak[1] - last_sd) / last_sd) < tol):
                last_centre = gauss_peak[0]
                last_sd = gauss_peak[1]
                converged = True
                break

            last_centre = gauss_peak[0]
            last_sd = gauss_peak[1]

        if not converged:
            warnings.warn('specular ridge search did not work properly'
                          ' using last known centre', RuntimeWarning)
            if manual_beam_find is not None:
                ret = manual_beam_find(detector[j], detector_sd[j])
                beam_centre[j], beam_sd[j], lopx[j], hipx[j], bp = ret
                background_mask[j, :, bp[0]] = True

                # don't assign to beam_centre, etc, at the end of this loop
                continue

        beam_centre[j] = last_centre
        beam_sd[j] = np.abs(last_sd)
        lp, hp, bp = fore_back_region(beam_centre[j], beam_sd[j])
        lopx[j] = lp
        hipx[j] = hp

        # bp are the background pixels. Clip to the range of the detector
        bp = np.clip(bp[0], 0, np.size(detector, 2) - 1)
        bp = np.unique(bp)
        background_mask[j, :, bp] = True

    # the foreground region needs to be totally contained within the
    # detector
    if (lopx < 0).any():
        raise ValueError("The foreground region for one of the detector"
                         " images extends below pixel 0.")
    if (hipx > np.size(detector, 2) - 1).any():
        raise ValueError("The foreground region for one of the detector"
                         " images extends above the largest detector"
                         " pixel.")

    return beam_centre, beam_sd, lopx, hipx, background_mask


def fore_back_region(beam_centre, beam_sd):
    """
    Calculates the fore and background regions based on the beam centre and
    width

    Parameters
    ----------
    beam_centre : float
        beam_centre
    beam_sd : float
        beam width (standard deviation)

    Returns
    -------
    lopx, hipx, background_pixels: float, float, list
        Lowest pixel of foreground region
        Highest pixel of foreground region
        Pixels that are in the background region
        Each of these should have `len(lopx) == len(beam_centre)`
    """
    _b_centre = np.array(beam_centre)
    _b_sd = np.array(beam_sd)

    lopx = np.floor(_b_centre - _b_sd * EXTENT_MULT).astype('int')
    hipx = np.ceil(_b_centre + _b_sd * EXTENT_MULT).astype('int')

    background_pixels = []

    # limit of background regions
    # from refnx.reduce.platypusnexus
    y1 = np.atleast_1d(
        np.round(lopx - PIXEL_OFFSET).astype('int'))
    y0 = np.atleast_1d(
        np.round(lopx - PIXEL_OFFSET - (EXTENT_MULT * _b_sd)).astype('int'))

    y2 = np.atleast_1d(
        np.round(hipx + PIXEL_OFFSET).astype('int'))
    y3 = np.atleast_1d(
        np.round(hipx + PIXEL_OFFSET + (EXTENT_MULT * _b_sd)).astype('int'))

    # now generate background pixels
    for i in range(np.size(y0)):
        background_pixels.append(np.r_[np.arange(y0[i], y1[i] + 1),
                                       np.arange(y2[i], y3[i] + 1)])

    return lopx, hipx, background_pixels


def correct_for_gravity(detector, detector_sd, lamda, coll_distance,
                        sample_det, lo_wavelength, hi_wavelength,
                        theta=0):
    """
    Returns a gravity corrected yt plot, given the data, its associated errors,
    the wavelength corresponding to each of the time bins, and the trajectory
    of the neutrons. Low lambda and high Lambda are wavelength cutoffs to
    ignore.

    Parameters
    ----------
    detector : np.ndarray
        Detector image. Has shape (N, T, Y)
    detector_sd : np.ndarray
        Standard deviations of detector image
    lamda : np.ndarray
        Wavelengths corresponding to the detector image, has shape (N, T)
    coll_distance : float
        Collimation distance between slits, mm
    sample_det : float
        Sample - detector distance, mm
    lo_wavelength : float
        Low wavelength cut off, Angstrom
    hi_wavelength : float
        High wavelength cutoff, Angstrom
    theta : float
        Angle between second collimation slit, first collimation slit, and
        horizontal

    Returns
    -------
    corrected_data, corrected_data_sd, m_gravcorrcoefs :
        np.ndarray, np.ndarray, np.ndarray

        Corrected image. This is a theoretical prediction where the spectral
        ridge is for each wavelength.  This will be used to calculate the
        actual angle of incidence in the reduction process.

    """
    x_init = np.arange((np.size(detector, axis=2) + 1) * 1.) - 0.5

    m_gravcorrcoefs = np.zeros((np.size(detector, 0)), dtype='float64')

    corrected_data = np.zeros_like(detector)
    corrected_data_sd = np.zeros_like(detector)

    for spec in range(np.size(detector, 0)):
        neutron_speeds = general.wavelength_velocity(lamda[spec])
        trajectories = find_trajectory(coll_distance / 1000.,
                                       theta,
                                       neutron_speeds)
        travel_distance = (coll_distance + sample_det[spec]) / 1000.

        # centres(t,)
        # TODO, don't use centroids, use Gaussian peak
        centroids = np.apply_along_axis(centroid, 1, detector[spec])
        lopx = np.searchsorted(lamda[spec], lo_wavelength)
        hipx = np.searchsorted(lamda[spec], hi_wavelength)

        def f(tru_centre):
            deflections = y_deflection(trajectories[lopx: hipx],
                                       neutron_speeds[lopx: hipx],
                                       travel_distance)

            model = 1000. * deflections / Y_PIXEL_SPACING + tru_centre
            diff = model - centroids[lopx: hipx, 0]
            diff = diff[~np.isnan(diff)]
            return diff

        # find the beam centre for an infinitely fast neutron
        x0 = np.array([np.nanmean(centroids[lopx: hipx, 0])])
        res = leastsq(f, x0)
        m_gravcorrcoefs[spec] = res[0][0]

        total_deflection = 1000. * y_deflection(trajectories,
                                                neutron_speeds,
                                                travel_distance)
        total_deflection /= Y_PIXEL_SPACING

        x_rebin = x_init.T + total_deflection[:, np.newaxis]
        for wavelength in range(np.size(detector, axis=1)):
            output = rebin(x_init,
                           detector[spec, wavelength],
                           x_rebin[wavelength],
                           y1_sd=detector_sd[spec, wavelength])

            corrected_data[spec, wavelength] = output[0]
            corrected_data_sd[spec, wavelength] = output[1]

    return corrected_data, corrected_data_sd, m_gravcorrcoefs


def calculate_wavelength_bins(lo_wavelength, hi_wavelength, rebin_percent):
    """
    Calculates optimal logarithmically spaced wavelength histogram bins. The
    bins are equal size in log10 space, but they may not be exactly be
    `rebin_percent` in size. The limits would have to change slightly for that.

    Parameters
    ----------
    lo_wavelength : float
        Low wavelength cutoff
    hi_wavelength : float
        High wavelength cutoff
    rebin_percent : float
        Rebinning percentage

    Returns
    -------
    wavelength_bins : np.ndarray
    """
    frac = (rebin_percent / 100.) + 1
    lowspac = rebin_percent / 100. * lo_wavelength
    hispac = rebin_percent / 100. * hi_wavelength

    lowl = lo_wavelength - lowspac / 2.
    hil = hi_wavelength + hispac / 2.
    num_steps = int(np.floor(np.log10(hil / lowl) / np.log10(frac)) + 1)
    rebinning = np.logspace(np.log10(lowl), np.log10(hil), num=num_steps)
    return rebinning


def accumulate_HDF_files(files):
    r"""
    Accumulates HDF files together, writing an accumulated file in the current
    directory. The accumulated datafile is written in the current directory
    (os.getcwd()) and has a filename based on the first file, prepended by
    'ADD\_'. For example, if the first file is PLP0000708.nx.hdf then the
    accumulated file is ADD_PLP0000708.nx.hdf.

    Parameters
    ----------
    files : list
        Strings specifying NeXUS filenames to be added together.

    """
    # don't do anything if no files were supplied.
    if not len(files):
        return None

    # the first file is the "master file", lets copy it.
    file = files[0]

    pth = _check_HDF_file(file)
    if not pth:
        raise ValueError('All files must refer to an hdf5 file')

    new_name = 'ADD_' + os.path.basename(pth)

    shutil.copy(pth,
                os.path.join(os.getcwd(), new_name))

    master_file = os.path.join(os.getcwd(), new_name)
    with h5py.File(master_file, 'r+') as h5master:
        # now go through each file and accumulate numbers:
        for file in files[1:]:
            pth = _check_HDF_file(file)
            h5data = h5py.File(pth, 'r')

            h5master['entry1/data/hmm'][0] += \
                h5data['entry1/data/hmm'][0]
            h5master['entry1/monitor/bm1_counts'][0] += \
                h5data['entry1/monitor/bm1_counts'][0]
            h5master['entry1/instrument/detector/total_counts'][0] += \
                h5data['entry1/instrument/detector/total_counts'][0]
            h5master['entry1/instrument/detector/time'][0] += \
                h5data['entry1/instrument/detector/time'][0]
            h5master.flush()

            h5data.close()


def _check_HDF_file(h5data):
    # If a file is an HDF5 file, then return the filename.
    # otherwise return False
    if type(h5data) == h5py.File:
        return h5data.filename
    else:
        with h5py.File(h5data, 'r') as h5data:
            if type(h5data) == h5py.File:
                return h5data.filename

    return False


@contextmanager
def _possibly_open_hdf_file(f, mode='r'):
    """
    Context manager for hdf5 files.

    Parameters
    ----------
    f : file-like or str
        If `f` is a file, then yield the file. If `f` is a str then open the
        file and yield the newly opened file.
        On leaving this context manager the file is closed, if it was opened
        by this context manager (i.e. `f` was a string).
    mode : str, optional
        mode is an optional string that specifies the mode in which the file
        is opened.

    Yields
    ------
    g : file-like
        On leaving the context manager the file is closed, if it was opened by
        this context manager.
    """
    close_file = False
    if type(f) == h5py.File:
        g = f
    else:
        g = h5py.File(f, mode)
        close_file = True
    yield g
    if close_file:
        g.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process some Platypus NeXUS'
                                                 'files to produce their TOF '
                                                 'spectra.')

    parser.add_argument('file_list', metavar='N', type=int, nargs='+',
                        help='integer file numbers')

    parser.add_argument('-b', '--bdir', type=str,
                        help='define the location to find the nexus files')

    parser.add_argument('-d', '--direct', action='store_true', default=False,
                        help='is the file a direct beam?')

    parser.add_argument('-r', '--rebin', type=float,
                        help='rebin percentage for the wavelength -1<rebin<10',
                        default=1)

    parser.add_argument('-ll', '--lolambda', type=float,
                        help='lo wavelength cutoff for the rebinning',
                        default=2.5)

    parser.add_argument('-hl', '--hilambda', type=float,
                        help='lo wavelength cutoff for the rebinning',
                        default=19.)

    parser.add_argument('-i', '--integrate', type=int,
                        help='-1 to integrate all spectra, otherwise enter the'
                             ' spectrum number.', default=-1)
    args = parser.parse_args()

    for file in args.file_list:
        fname = 'PLP%07d.nx.hdf' % file
        path = os.path.join(args.bdir, fname)
        try:
            a = PlatypusNexus(path)
            a.process(lo_wavelength=args.lolambda,
                      hi_wavelength=args.hilambda,
                      direct=args.direct,
                      rebin_percent=args.rebin,
                      integrate=args.integrate)

            fname = 'PLP%07d.spectrum' % file
            out_fname = os.path.join(args.bdir, fname)

            integrate = args.integrate
            if args.integrate < 0:
                integrate = 0

            a.write_spectrum_dat(out_fname, scanpoint=integrate)

        except IOError:
            print("Couldn't find file: %d.  Use --basedir option" % file)



"""
Testing rebin histogram values.
"""

import numpy as np
from numpy.random import uniform
from numpy.testing import (assert_allclose, assert_equal)
import uncertainties.unumpy as unp
import refnx.reduce.rebin as rebin


class TestRebin(object):
    def setup_method(self):
        pass

    # -----------------------------------------------------------------------#
    #  Tests for piecewise continuous rebinning
    # -----------------------------------------------------------------------#
    def test_x2_same_as_x1(self):
        """
        x2 same as x1
        """
        # old size
        m = 6

        # new size
        n = 6

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(0., 1., n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # rebin
        y_new = rebin.rebin(x_old, y_old, x_new)

        assert_allclose(y_new, y_old)

    def test_x2_surrounds_x1(self):
        """
        x2 range surrounds x1 range
        """
        # old size
        m = 2

        # new size
        n = 3

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(-0.1, 1.2, n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # rebin
        y_new = rebin.rebin(x_old, y_old, x_new)

        # compute answer here to check rebin
        y_old_ave = y_old / np.diff(x_old)
        y_new_here = [y_old_ave[0] * (x_new[1] - 0.),
                      y_old_ave[0] * (x_old[1] - x_new[1]) +
                      y_old_ave[1] * (x_new[2] - x_old[1]),
                      y_old_ave[1] * (x_old[-1] - x_new[-2])]

        assert_allclose(y_new, y_new_here)
        assert_allclose(y_new.sum(), y_old.sum())

    def test_x2_surrounds_x1_2(self):
        """
        x2 has some bins that span several x1 bins
        Also tests uncertainty propagation. Values calculated using
        original jhykes piecewise constant code
        """
        # old size
        m = 10

        # new size
        n = 3

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(-0.1, 1.2, n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # with uncertainties
        np.random.seed(1)
        y_old_sd = 0.1 * y_old * uniform((m,))

        # rebin
        y_new, y_new_sd = rebin.rebin(x_old,
                                      y_old,
                                      x_new,
                                      y1_sd=y_old_sd)

        # compute answer here to check rebin
        y_new_here = np.array([14.99807911, 44.14135692, 13.99807911])
        y_new_here_sd = np.array([5.381524308729351,
                                  12.73174109312833,
                                  5.345145324353735])

        assert_allclose(y_new, y_new_here)
        assert_allclose(y_new_sd, y_new_here_sd)
        assert_allclose(y_new.sum(), y_old.sum())

    def test_x2_lower_than_x1(self):
        """
        x2 range is completely lower than x1 range
        """
        # old size
        m = 2

        # new size
        n = 3

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(-0.2, -0.0, n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # rebin
        y_new = rebin.rebin(x_old, y_old, x_new)

        assert_allclose(y_new, [0., 0., 0.])
        assert_allclose(y_new.sum(), 0.)

    def test_x2_above_x1(self):
        """
        x2 range is completely above x1 range
        """
        # old size
        m = 20

        # new size
        n = 30

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(1.2, 10., n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # rebin
        y_new = rebin.rebin(x_old, y_old, x_new)

        assert_allclose(y_new, np.zeros((n,)))
        assert_allclose(y_new.sum(), 0.)

    def test_x2_in_x1(self):
        """
        x2 only has one bin, and it is surrounded by x1 range
        """
        # old size
        m = 4

        # new size
        n = 1

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(0.3, 0.65, n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # rebin
        y_new = rebin.rebin(x_old, y_old, x_new)

        # compute answer here to check rebin
        y_old_ave = y_old / np.diff(x_old)
        y_new_here = (y_old_ave[1] * (x_old[2] - x_new[0]) +
                      y_old_ave[2] * (x_new[1] - x_old[2]))

        assert_allclose(y_new, y_new_here)

    def test_y1_uncertainties(self):
        """
        x2 range surrounds x1 range, y1 has uncertainties
        """
        # old size
        m = 2

        # new size
        n = 3

        # bin edges
        x_old = np.linspace(0., 1., m + 1)
        x_new = np.linspace(-0.1, 1.2, n + 1)

        # some arbitrary distribution
        y_old = 1. + np.sin(x_old[:-1] * np.pi) / np.diff(x_old)

        # with uncertainties
        y_old = unp.uarray(y_old, 0.1 * y_old * uniform((m,)))

        # rebin
        y_new, y_new_sd = rebin.rebin(x_old,
                                      unp.nominal_values(y_old),
                                      x_new,
                                      y1_sd=unp.std_devs(y_old))

        # compute answer here to check rebin
        y_old_ave = y_old / np.diff(x_old)
        y_new_here = np.array(
            [y_old_ave[0] * (x_new[1] - 0.),
             y_old_ave[0] * (x_old[1] - x_new[1]) +
             y_old_ave[1] * (x_new[2] - x_old[1]),
             y_old_ave[1] * (x_old[-1] - x_new[-2])])

        # mean or nominal value comparison
        assert_allclose(y_new,
                        unp.nominal_values(y_new_here))

        # mean or nominal value comparison
        assert_allclose(y_new_sd,
                        unp.std_devs(y_new_here))

        assert_allclose(y_new.sum(),
                        unp.nominal_values(y_new_here).sum())

    def test_rebinND(self):
        input = np.arange(24).reshape(4, 3, 2)
        x1 = np.arange(5)
        x2 = np.array([0, 1, 2.5, 4])
        y1 = np.arange(4)
        y2 = np.array([0.5, 1.5])
        output = rebin.rebinND(input, (0, 1), (x1, y1), (x2, y2))
        res = np.array([[[1., 2.]],
                        [[13.5, 15.]],
                        [[25.5, 27.]]])

        assert_equal(res, output)



# -*- coding: utf-8 -*-
"""
Unpack event files
"""
import numpy as np
HAVE_CEVENTS = False
try:
    from refnx.reduce._cevent import _cevents
    HAVE_CEVENTS = True
except ImportError:
    pass

ii32 = np.iinfo(np.int32)


def process_event_stream(events, frames, t_bins, y_bins, x_bins):
    """
    Processes the event mode dataset into a histogrammed detector image.

    Parameters
    ----------
    events : tuple(np.ndarray, np.ndarray, np.ndarray, np.ndarray)
        the 4-tuple of events (F, T, Y, X)
    frames : sequence of array-like
        each array in the sequence specifies which frames should be included
    t_bins : array_like
        specifies the time bins required in the image
    y_bins : array_like
        specifies the y bins required in the image
    x_bins : array_like
        specifies the x bins required in the image
    Returns
    -------
    detector, frame_count : np.ndarray, np.ndarray
        The new detector images and the number of frames within each image.
        `detector` has shape
        `(len(frames), len(t_bins) - 1, len(y_bins) - 1, len(x_bins) - 1)`.
        `frame_count` says how many frames went into making each frame in the
        detector image

    Notes
    -----
    Every entry in `frames` is clipped to 0 and the maximum frame number in the
    events. Thus, if
    frames = [[-2, -1, 0, 1, 2, 3]], and the maximum frame number is 2, then
    only the 0, 1, 2 frames are included.

    Examples
    ---------
    >>> from refnx.reduce.event import process_event_stream, events
    >>> d = process_event_stream(e, [np.linspace(0, 2308585, 2308586)],
    ...     [0, 40000], np.linspace(110.5, -110.5, 222),
    ...     np.linspace(210.5, -210.5, 422))
    """
    max_frame = max(events[0])

    t_events = np.asarray(events).T

    localxbins = np.array(x_bins)
    localybins = np.array(y_bins)
    localtbins = np.sort(np.array(t_bins))
    reversed_x, reversed_y = False, False

    if localxbins[0] > localxbins[-1]:
        localxbins = localxbins[::-1]
        reversed_x = True

    if localybins[0] > localybins[-1]:
        localybins = localybins[::-1]
        reversed_y = True

    # create an (N, T, Y, X) detector image
    detector = np.zeros((len(frames),
                         len(t_bins) - 1,
                         len(y_bins) - 1,
                         len(x_bins) - 1),
                        dtype=np.uint64)
    frame_count = np.zeros(len(frames))

    for i, frame in enumerate(frames):
        frame_numbers = np.unique(np.clip(np.asarray(frame), 0, max_frame))
        frame_count[i] = frame_numbers.size

        frames_with_events = set(frame_numbers).intersection(t_events[:, 0])

        frame_numbers = list(frames_with_events)
        frame_numbers.sort()

        left = np.searchsorted(t_events[:, 0], frame_numbers)
        right = np.searchsorted(t_events[:, 0], frame_numbers, side='right')
        idxs = np.concatenate([np.arange(l, r) for l, r in zip(left, right)])

        filtered_events = t_events[idxs]

        detector[i], edge = np.histogramdd(filtered_events[:, 1:],
                                           bins=(localtbins,
                                                 localybins,
                                                 localxbins))
    if reversed_x:
        detector = detector[:, :, :, ::-1]

    if reversed_y:
        detector = detector[:, :, ::-1, :]

    return detector, frame_count


def framebins_to_frames(frame_bins):
    if frame_bins is None:
        return None

    t_frame_bins = np.asarray(frame_bins)

    frames = []
    for idx in range(t_frame_bins.size - 1):
        frames.append(np.arange(t_frame_bins[idx],
                                t_frame_bins[idx + 1], dtype=np.uint64))

    return frames


def events(f, end_last_event=127, max_frames=None):
    """
    Unpacks event data from packedbinary format for the ANSTO Platypus
    instrument

    Parameters
    ----------

    f : file-like or str
        The file to read the data from. If `f` is not file-like then f is
        assumed to be a path pointing to the event file.
    end_last_event : uint
        The reading of event data starts from `end_last_event + 1`. The default
        of 127 corresponds to a file header that is 128 bytes long.
    max_frames : None, int
        Stop reading the event file when have read this many frames. If `None`
        then read all frames

    Returns
    -------
    (f_events, t_events, y_events, x_events), end_last_event:
        x_events, y_events, t_events and f_events are numpy arrays containing
        the events. end_last_event is a byte offset to the end of the last
        successful event read from the file. Use this value to extract more
        events from the same file at a future date.
    """
    if HAVE_CEVENTS:
        return _cevents(f, end_last_event=end_last_event,
                        max_frames=max_frames)
    else:
        return _events(f, end_last_event=end_last_event,
                       max_frames=max_frames)


def _events(f, end_last_event=127, max_frames=None):
    """
    Unpacks event data from packedbinary format for the ANSTO Platypus
    instrument

    Parameters
    ----------

    f : file-like or str
        The file to read the data from. If `f` is not file-like then f is
        assumed to be a path pointing to the event file.
    end_last_event : uint
        The reading of event data starts from `end_last_event + 1`. The default
        of 127 corresponds to a file header that is 128 bytes long.
    max_frames : None, int
        Stop reading the event file when have read this many frames. If `None`
        then read all frames.

    Returns
    -------
    (f_events, t_events, y_events, x_events), end_last_event:
        x_events, y_events, t_events and f_events are numpy arrays containing
        the events. end_last_event is a byte offset to the end of the last
        successful event read from the file. Use this value to extract more
        events from the same file at a future date.
    """
    if max_frames is None:
        max_frames = ii32.max

    fi = f
    auto_f = None
    if not hasattr(fi, 'read'):
        auto_f = open(f, 'rb')
        fi = auto_f

    frame_number = -1
    dt = 0
    t = 0
    x = -0
    y = -0

    x_events = np.array((), dtype='int32')
    y_events = np.array((), dtype='int32')
    t_events = np.array((), dtype='uint32')
    f_events = np.array((), dtype='int32')

    bufsize = 32768

    while True and frame_number < max_frames:
        x_neutrons = []
        y_neutrons = []
        t_neutrons = []
        f_neutrons = []

        fi.seek(end_last_event + 1)
        buf = fi.read(bufsize)

        filepos = end_last_event + 1

        if not len(buf):
            break

        buf = bytearray(buf)
        state = 0

        for i, c in enumerate(buf):
            if state == 0:
                x = c
                state += 1
            elif state == 1:
                x |= (c & 0x3) * 256

                if x & 0x200:
                    x = - (0x100000000 - (x | 0xFFFFFC00))
                y = int(c / 4)
                state += 1
            else:
                if state == 2:
                    y |= (c & 0xF) * 64

                    if y & 0x200:
                        y = -(0x100000000 - (y | 0xFFFFFC00))
                event_ended = ((c & 0xC0) != 0xC0 or state >= 7)

                if not event_ended:
                    c &= 0x3F
                if state == 2:
                    dt = c >> 4
                else:
                    dt |= c << 2 + 6 * (state - 3)

                if not event_ended:
                    state += 1
                else:
                    # "state", state, event_ended, x, y, frame_number, t, dt
                    state = 0
                    end_last_event = filepos + i
                    if x == 0 and y == 0 and dt == 0xFFFFFFFF:
                        t = 0
                        frame_number += 1
                        if frame_number == max_frames:
                            break
                    else:
                        t += dt
                        if frame_number == -1:
                            return None
                        x_neutrons.append(x)
                        y_neutrons.append(y)
                        t_neutrons.append(t)
                        f_neutrons.append(frame_number)

        if x_neutrons:
            x_events = np.append(x_events, x_neutrons)
            y_events = np.append(y_events, y_neutrons)
            t_events = np.append(t_events, t_neutrons)
            f_events = np.append(f_events, f_neutrons)

    t_events //= 1000

    if auto_f:
        auto_f.close()

    return (f_events, t_events, y_events, x_events), end_last_event



import os.path
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import pyqtSlot
import periodictable as pt
import pyparsing
import numpy as np


UI_LOCATION = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'ui')


class SLDcalculatorView(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super(SLDcalculatorView, self).__init__(parent)

        self.ui = uic.loadUi(
            os.path.join(UI_LOCATION,
                         'SLDcalculator.ui'),
            self)
        self.last_formula = 'H[2]2O'

    def calculate(self):
        try:
            formula = pt.formula(self.ui.chemical_formula.text(),)

        except (pyparsing.ParseException, TypeError, ValueError):
            return

        if not len(formula.atoms):
            return

        density = self.ui.mass_density.value()
        molecular_volume = self.ui.molecular_volume.value()
        neutron_wavelength = self.ui.neutron_wavelength.value()
        xray_energy = self.ui.xray_energy.value()

        if self.ui.use_volume.isChecked():
            density = formula.molecular_mass / formula.volume(
                a=molecular_volume,
                b=1,
                c=1)
            self.ui.mass_density.setValue(density)

        elif self.ui.use_density.isChecked():
            try:
                volume = (formula.mass / density /
                          pt.constants.avogadro_number * 1e24)
            except ZeroDivisionError:
                volume = np.nan
            self.ui.molecular_volume.setValue(volume)

        try:
            real, imag, mu = pt.neutron_sld(formula,
                                            density=density,
                                            wavelength=neutron_wavelength)

            self.ui.neutron_SLD.setText(
                '%.6g' %
                real +
                ' + ' +
                '%.6g' %
                imag +
                'j')
        except Exception:
            self.ui.neutron_SLD.setText('NaN')

        try:
            real, imag = pt.xray_sld(formula,
                                     density=density,
                                     energy=xray_energy)

            self.ui.xray_SLD.setText('%.6g' % real + ' + ' +
                                     '%.6g' % imag + 'j')
        except Exception:
            self.ui.xray_SLD.setText('NaN')
            # sometimes the X-ray and neutron SLD calc can fail, if there are
            # no scattering factors

    @pyqtSlot(float)
    def on_mass_density_valueChanged(self, arg_1):
        self.calculate()

    @pyqtSlot(float)
    def on_molecular_volume_valueChanged(self, arg_1):
        self.calculate()

    @pyqtSlot(float)
    def on_neutron_wavelength_valueChanged(self, arg_1):
        self.calculate()

    @pyqtSlot(float)
    def on_xray_energy_valueChanged(self, arg_1):
        self.calculate()

    @pyqtSlot(str)
    def on_chemical_formula_textChanged(self, arg_1):
        self.calculate()



from multiprocessing import Pool, get_context
import warnings as _warnings
import os as _os
import sys as _sys
import functools
from tempfile import mkdtemp
from collections.abc import Iterable
from contextlib import contextmanager
from inspect import getfullargspec as _getargspecf


def preserve_cwd(function):
    """
    Ensures that the original working directory is kept when exiting a function

    Parameters
    ----------
    function : callable

    """
    @functools.wraps(function)
    def decorator(*args, **kwargs):
        cwd = _os.getcwd()
        try:
            return function(*args, **kwargs)
        finally:
            _os.chdir(cwd)
    return decorator


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        # Handle mkdtemp raising an exception
        self.name = None
        self.name = mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                print("ERROR: {!r} while cleaning up {!r}".format(ex, self,),
                      file=_sys.stderr)
                return
            self._closed = True
            if _warn:
                # ResourceWarning
                self._warn("ResourceWarning: Implicitly cleaning"
                           " up {!r}".format(self))

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass


def flatten(l):
    """
    Flatten a nested sequence.

    Parameters
    ----------
    l : sequence
        The sequence to flatten

    Returns
    -------
    el : generator
        yields flattened sequences from l
    """
    for el in l:
        if (isinstance(el, Iterable) and
                not isinstance(el, (str, bytes))):
            yield from flatten(el)
        else:
            yield el


def unique(seq, idfun=id):
    """
    List of unique values in sequence (by object id, not by value).
    Ordering is preserved.

    Parameters
    ----------
    seq : sequence
    idfun : callable

    Returns
    -------
    p : generator
        yields unique values from l

    Notes
    -----
    Because this function works on object id (by default), it won't work
    for looking at the number of unique values in a numpy array - all
    the entries have different object id's.
    """
    seen = {}
    for item in seq:
        marker = idfun(item)
        if marker not in seen:
            seen[marker] = 1
            yield item


@contextmanager
def possibly_open_file(f, mode='wb'):
    """
    Context manager for files.

    Parameters
    ----------
    f : file-like or str
        If `f` is a file, then yield the file. If `f` is a str then open the
        file and yield the newly opened file.
        On leaving this context manager the file is closed, if it was opened
        by this context manager (i.e. `f` was a string).
    mode : str, optional
        mode is an optional string that specifies the mode in which the file
        is opened.

    Yields
    ------
    g : file-like
        On leaving the context manager the file is closed, if it was opened by
        this context manager.
    """
    close_file = False
    if (hasattr(f, 'read') and hasattr(f, 'write')) or f is None:
        g = f
    else:
        g = open(f, mode)
        close_file = True
    yield g
    if close_file:
        g.close()


class MapWrapper(object):
    """
    Parallelisation wrapper for working with map-like callables, such as
    `multiprocessing.Pool.map`.

    Parameters
    ----------
    pool : int or map-like callable
        If `pool` is an integer, then it specifies the number of threads to
        use for parallelization. If ``int(pool) == 1``, then no parallel
        processing is used and the map builtin is used.
        If ``pool == -1``, then the pool will utilise all available CPUs.
        If `pool` is a map-like callable that follows the same
        calling sequence as the built-in map function, then this callable is
        used for parallelisation.
    context : None, {'spawn', 'fork', 'forkserver'}

    """
    def __init__(self, pool=-1, context=None):
        self.pool = None
        self._mapfunc = map
        self._own_pool = False

        ctx = get_context(context)

        if callable(pool):
            self.pool = pool
            self._mapfunc = self.pool
        else:
            # user supplies a number
            if int(pool) == -1:
                # use as many processors as possible
                self.pool = ctx.Pool()
                self._mapfunc = self.pool.map
                self._own_pool = True
            elif int(pool) in [0, 1]:
                pass
            elif int(pool) > 1:
                # use the number of processors requested
                self.pool = ctx.Pool(processes=int(pool))
                self._mapfunc = self.pool.map
                self._own_pool = True

    def __enter__(self):
        return self

    def __del__(self):
        self.close()
        self.terminate()

    def terminate(self):
        if self._own_pool:
            self.pool.terminate()

    def join(self):
        if self._own_pool:
            self.pool.join()

    def close(self):
        if self._own_pool:
            self.pool.close()

    def __exit__(self, exc_type, exc_value, traceback):
        if self._own_pool:
            self.pool.close()
            self.pool.terminate()

    def __call__(self, func, iterable):
        # only accept one iterable because that's all Pool.map accepts
        try:
            return self._mapfunc(func, iterable)
        except TypeError:
            # wrong number of arguments
            raise TypeError("The map-like callable must be of the"
                            " form f(func, iterable)")

    def map(self, func, iterable):
        return self(func, iterable)


def getargspec(f):
    return _getargspecf(f)



from refnx.analysis.bounds import Interval, PDF, Bounds
from refnx.analysis.parameter import (Parameter, Parameters, is_parameters,
                                      is_parameter, possibly_create_parameter)
from refnx.analysis.objective import (Objective, BaseObjective,
                                      GlobalObjective, Transform,
                                      pymc_objective)
from refnx.analysis.curvefitter import (CurveFitter, MCMCResult, process_chain,
                                        load_chain, autocorrelation_chain)
from refnx._lib.emcee.autocorr import integrated_time
from refnx.analysis.model import Model, fitfunc


__all__ = [s for s in dir() if not s.startswith('_')]


from refnx._lib._testutils import PytestTester

test = PytestTester(__name__)
del PytestTester
